"""This file contains prompts for CritiqueAndRefineConvert operations."""

### CRITIQUE PROMPT AND CRITERIA ###
BASE_CRITIQUE_PROMPT = """You are a helpful assistant tasked with critiquing the output of a model based on a given prompt.
Below is the original user prompt used to generate the output:

ORIGINAL PROMPT:
<<original-prompt-placeholder>>

Here is the output generated by the model:

OUTPUT:
{original_output}

Your task is to critique the output based on the following:
{critique_criteria}

{finish_instruction}
"""

MAP_CRITIQUE_CRITERIA = """1. Does the JSON object adhere to the required format? Highlight any structural issues.
2. Are the values of the output fields accurate based on the provided context? If any fields are incorrect or missing, provide specific examples.
3. Are there any logical errors in reasoning used to derive the output? Provide detailed feedback on potential mistakes.
"""
FILTER_CRITIQUE_CRITERIA = """1. Does the output adhere to the required TRUE/FALSE format? Highlight any issues.
2. Is the TRUE/FALSE determination accurate based on the provided context? If there is evidence for an incorrect determination, provide specific reasons why.
3. Are there any logical errors in reasoning used to derive the TRUE/FALSE determination? Provide detailed feedback on potential mistakes.
"""

MAP_CRITIQUE_FINISH_INSTRUCTION = """Finish your critique with actionable recommendations for improving the JSON object."""
FILTER_CRITIQUE_FINISH_INSTRUCTION = """Finish your critique with actionable recommendations for improving the model's reasoning and answer."""

### REFINEMENT PROMPT AND CRITERIA ###
BASE_REFINEMENT_PROMPT = """You are a helpful assistant tasked with refining the output of a model based on a critique.
Below is the original user prompt used to generate the output:

ORIGINAL PROMPT:
<<original-prompt-placeholder>>

Here is the original output generated by the model:

OUTPUT:
{original_output}

Here is the critique of the output:

CRITIQUE:
{critique_output}

Your task is to refine the original output to address the critique. Ensure that:
{refinement_criteria}

{finish_instruction}
"""

MAP_REFINEMENT_CRITERIA = """1. The answer adheres to the required JSON format specified in the original prompt.
2. Correctly derives all values for the output fields based on the provided context.
3. Resolves any logical errors identified in the critique.
"""
FILTER_REFINEMENT_CRITERIA = """1. The answer adheres to the required TRUE/FALSE format specified in the original prompt.
2. Correctly derives the final TRUE/FALSE determination based on the provided context.
3. Resolves any logical errors identified in the critique.
"""

MAP_REFINEMENT_FINISH_INSTRUCTION = """Return the refined JSON object as your final answer."""
FILTER_REFINEMENT_FINISH_INSTRUCTION = """Return the final TRUE/FALSE answer."""



from __future__ import annotations

import os
import re
from abc import ABC
from typing import Callable

from pydantic import BaseModel
from smolagents import CodeAgent, LiteLLMModel

from palimpzest.core.data import context_manager
from palimpzest.core.data.dataset import Dataset
from palimpzest.core.lib.schemas import create_schema_from_fields, union_schemas
from palimpzest.query.operators.logical import ComputeOperator, ContextScan, LogicalOperator, SearchOperator
from palimpzest.utils.hash_helpers import hash_for_id

PZ_INSTRUCTION = """\n\nYou are a CodeAgent who is a specialist at writing declarative AI programs with the Palimpzest (PZ) library.

Palimpzest is a programming framework which provides you with **semantic operators** (e.g. semantic maps, semantic filters, etc.)
which are like their traditional counterparts, except they can execute instructions provided in natural language.

For example, if you wanted to write a program to extract the title and abstract from a directory of papers,
you could write the following in PZ:
```
import palimpzest as pz
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

# Define columns for semantic map (sem_map) operation; each column is specified
# with a dictionary containing the following keys:
# - "name": the name of the field to compute
# - "type": the type of the field to compute
# - "description": the natural language description of the field
paper_cols = [
    {"name": "title", "type": str, "description": "the title of the paper"},
    {"name": "abstract", "type": str, "description": "the paper's abstract"},
]

# construct the data processing pipeline with PZ
ds = pz.TextFileDataset(id="papers", path="path/to/papers")
ds = ds.sem_map(cols)

# optimize and execute the PZ program
validator = pz.Validator()
config = pz.QueryProcessorConfig(
    policy=pz.MaxQuality(),
    execution_strategy="parallel",
    max_workers=20,
    progress=True,
)
output = ds.optimize_and_run(config=config, validator=validator)

# write the execution stats to json
output.execution_stats.to_json("pz_program_stats.json")

# write the output to a CSV and print the output CSV filepath so the user knows where to find it
output_filepath = "pz_program_output.csv"
output.to_df().to_csv(output_filepath, index=False)
print(f"Results at: {output_filepath}")
```

To initialize a dataset in PZ, simply provide the path to a directory to `pz.TextFileDirectory()`
(if your data contains text-based files). For example:
```
import palimpzest as pz
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

ds = pz.TextFileDataset(id="files", path="path/to/files")
```

Palimpzest has two primary **semantic operators** which you can use to construct data processing pipelines:
- sem_filter(predicate: str): executes a semantic filter specified by the natural language predicate on a given PZ dataset
- sem_map(cols: list[dict]): executes a semantic map to compute the `cols` on a given PZ dataset

As a second example, consider the following PZ program which filters for papers about batteries that are from MIT
and computes a summary for each one:
```
import palimpzest as pz
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

# construct the PZ program
ds = pz.TextFileDataset(id="papers", path="path/to/research-papers")
ds = ds.sem_filter("The paper is about batteries")
ds = ds.sem_filter("The paper is from MIT")
ds = ds.sem_map([{"name": "summary", "type": str, "description": "A summary of the paper"}])

# optimize and execute the PZ program
validator = pz.Validator()
config = pz.QueryProcessorConfig(
    policy=pz.MaxQuality(),
    execution_strategy="parallel",
    max_workers=20,
    progress=True,
)
output = ds.optimize_and_run(config=config, validator=validator)

# write the execution stats to json
output.execution_stats.to_json("pz_program_stats.json")

# write the output to a CSV and print the output CSV filepath so the user knows where to find it
output_filepath = "pz_program_output.csv"
output.to_df().to_csv(output_filepath, index=False)
print(f"Results at: {output_filepath}")
```

Be sure to always:
- execute your program using the `.optimize_and_run()` format shown above
- call `output.execution_stats.to_json("pz_program_stats.json")` to write execution statistics to disk
- write your output to CSV and print where you wrote it!
"""

class Context(Dataset, ABC):
    """
    The `Context` class is an abstract base class for root `Datasets` whose data is accessed
    via user-defined methods. Classes which inherit from this class must implement two methods:

    - `list_filepaths()`: which lists the files that the `Context` has access to.
    - `read_filepath(path: str)`: which reads the file corresponding to the given `path`.

    A `Context` is a special type of `Dataset` that represents a view over an underlying `Dataset`.
    Each `Context` has a `name` which uniquely identifies it, as well as a natural language `description`
    of the data / computation that the `Context` represents. Similar to `Dataset`s, `Context`s can be
    lazily transformed using functions such as `sem_filter`, `sem_map`, `sem_join`, etc., and they may
    be materialized or unmaterialized.
    """

    def __init__(
            self,
            id: str,
            description: str,
            operator: LogicalOperator,
            schema: type[BaseModel] | None = None,
            sources: list[Context] | Context | None = None,
            materialized: bool = False,
        ) -> None:
        """
        Constructor for the `Context` class.

        Args:
            id (`str`): a string identifier for the `Context`
            description (`str`): the description of the data contained within the `Context`
            operator (`LogicalOperator`): The `LogicalOperator` used to compute this `Context`.
            schema: (`type[BaseModel] | None`): The schema of this `Context`.
            sources (`list[Context] | Context | None`): The (list of) `Context(s)` which are input(s) to
                the operator used to compute this `Context`.
            materialized (`bool`): True if the `Context` has been computed, False otherwise
        """
        # set the description
        self._description = description

        # set the materialization status
        self._materialized = materialized

        # compute schema and call parent constructor
        if schema is None:
            schema = create_schema_from_fields([{"name": "context", "description": "The context", "type": str}])
        super().__init__(sources=sources, operator=operator, schema=schema, id=id)

        # set the tools associated with this Context
        self._tools = [getattr(self, attr) for attr in dir(self) if attr.startswith("tool_")]

        # add Context to ContextManager
        cm = context_manager.ContextManager()
        cm.add_context(self)

    @property
    def description(self) -> str:
        """The string containing all of the information computed for this `Context`"""
        return self._description

    @property
    def materialized(self) -> bool:
        """The boolean which specifies whether the `Context` has been computed or not"""
        return self._materialized

    @property
    def tools(self) -> list[Callable]:
        """The list of tools associated with this `Context`"""
        return self._tools

    def __str__(self) -> str:
        return f"Context(id={self.id}, description={self.description:20s}, materialized={self.materialized})"

    def set_description(self, description: str) -> None:
        """
        Update the context's description.
        """
        self._description = description

    def set_materialized(self, materialized: str) -> None:
        """
        Update the context's materialization status.
        """
        self._materialized = materialized

    def compute(self, instruction: str) -> Context:
        # construct new description and output schema
        new_id = hash_for_id(instruction)
        new_description = f"Parent Context ID: {self.id}\n\nThis Context is the result of computing the following instruction on the parent context.\n\nINSTRUCTION: {instruction}\n\n"
        inter_schema = create_schema_from_fields([{"name": f"result-{new_id}", "desc": "The result from computing the instruction on the input Context",  "type": str}])
        new_output_schema = union_schemas([self.schema, inter_schema])

        # construct logical operator
        operator = ComputeOperator(
            input_schema=self.schema,
            output_schema=new_output_schema,
            context_id=new_id,
            instruction=instruction,
        )        

        return Context(id=new_id, description=new_description, operator=operator, sources=[self], materialized=False)

    def search(self, search_query: str) -> Context:
        # construct new description and output schema
        new_id = hash_for_id(search_query)
        new_description = f"Parent Context ID: {self.id}\n\nThis Context is the result of searching the parent context for information related to the following query.\n\nSEARCH QUERY: {search_query}\n\n"

        # construct logical operator
        operator = SearchOperator(
            input_schema=self.schema,
            output_schema=self.schema,
            context_id=new_id,
            search_query=search_query,
        )

        return Context(id=new_id, description=new_description, operator=operator, sources=[self], materialized=False)

class TextFileContext(Context):
    def __init__(self, path: str, id: str, description: str) -> None:
        """
        Constructor for the `TextFileContext` class.

        Args:
            path (str): The path to the file
            id (str): a string identifier for the `Context`
            description (str): The description of the data contained within the `Context`
            kwargs (dict): Keyword arguments containing the `Context's` id and description.
        """
        # check that path is a valid file or directory
        assert os.path.isfile(path) or os.path.isdir(path), f"Path {path} is not a file nor a directory"

        # get list of filepaths
        self.filepaths = []
        if os.path.isfile(path):
            self.filepaths = [path]
        else:
            self.filepaths = []
            for root, _, files in os.walk(path):
                for file in files:
                    fp = os.path.join(root, file)
                    self.filepaths.append(fp)
            self.filepaths = sorted(self.filepaths)

        # call parent constructor to set id, operator, and schema
        schema = create_schema_from_fields([{"name": "context", "desc": "The context", "type": str}])
        super().__init__(
            id=id,
            description=description,
            operator=ContextScan(context=self, output_schema=schema),
            schema=schema,
            materialized=True,
        )
    def _check_filter_answer_text(self, answer_text: str) -> dict | None:
        """
        Return {"passed_operator": True} if and only if "true" is in the answer text.
        Return {"passed_operator": False} if and only if "false" is in the answer text.
        Otherwise, return None.
        """
        # NOTE: we may be able to eliminate this condition by specifying this JSON output in the prompt;
        # however, that would also need to coincide with a change to allow the parse_answer_fn to set "passed_operator"
        if "true" in answer_text.lower():
            return {"passed_operator": True}
        elif "false" in answer_text.lower():
            return {"passed_operator": False}
        elif "yes" in answer_text.lower():
            return {"passed_operator": True}

        return None

    def _parse_filter_answer(self, completion_text: str) -> dict[str, list]:
        """Extract the answer from the completion object for filter operations."""
        # if the model followed the default instructions, the completion text will place
        # its answer between "ANSWER:" and "---"
        regex = re.compile("answer:(.*?)---", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()
            field_answers = self._check_filter_answer_text(answer_text)
            if field_answers is not None:
                return field_answers

        # if the first regex didn't find an answer, try taking all the text after "ANSWER:"
        regex = re.compile("answer:(.*)", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()
            field_answers = self._check_filter_answer_text(answer_text)
            if field_answers is not None:
                return field_answers

        # finally, try taking all of the text; throw an exception if this doesn't work
        field_answers = self._check_filter_answer_text(completion_text)
        if field_answers is None:
            raise Exception(f"Could not parse answer from completion text: {completion_text}")

        return field_answers

    # def tool_list_filepaths(self) -> list[str]:
    #     """
    #     This tool returns the list of all of the filepaths which the `Context` has access to.

    #     Args:
    #         None
        
    #     Returns:
    #         list[str]: A list of file paths for all files in the `Context`.
    #     """
    #     return self.filepaths

    # def tool_read_filepath(self, path: str) -> str:
    #     """
    #     This tool takes a filepath (`path`) as input and returns the content of the file as a string.
    #     It handles both CSV files and html / regular text files. It does not handle images.

    #     Args:
    #         path (str): The path to the file to read.

    #     Returns:
    #         str: The content of the file as a string.
    #     """
    #     if path.endswith(".csv"):
    #         return pd.read_csv(path, encoding="ISO-8859-1").to_string(index=False)

    #     with open(path, encoding='utf-8') as file:
    #         content = file.read()

    #     return content

    def tool_execute_semantic_operators(self, instruction: str) -> str:
        """
        This tool takes an `instruction` as input and invokes an expert to write a semantic data processing pipeline
        to execute the instruction. The tool returns the path to a CSV file which contains the output of the pipeline.

        For example, the tool could be invoked as follows to extract the title and abstract from a dataset of research papers:
        ```
        instruction = "Write a program to extract the title and abstract from each research paper"
        result_csv_filepath = tool_execute_semantic_operators(instruction)
        ```

        Args:
            instruction: The instruction specifying the semantic data processing pipeline that you need to execute.

        Returns:
            str: the filepath to the CSV containing the output from running the data processing pipeline.
        """
        from smolagents import tool
        @tool
        def tool_list_filepaths() -> list[str]:
            """
            This tool returns the list of all of the filepaths which the `Context` has access to.

            NOTE: You may want to execute this before writing your PZ program to determine where the data lives.

            Args:
                None
            
            Returns:
                list[str]: A list of file paths for all files in the `Context`.
            """
            return self.filepaths

        agent = CodeAgent(
            model=LiteLLMModel(model_id="openai/o1", api_key=os.getenv("ANTHROPIC_API_KEY")),
            tools=[tool_list_filepaths],
            max_steps=20,
            planning_interval=4,
            add_base_tools=False,
            return_full_result=True,
            additional_authorized_imports=["dotenv", "json", "palimpzest", "pandas"],
            instructions=PZ_INSTRUCTION,
        )
        result = agent.run(instruction)
        response = result.output

        return response



from __future__ import annotations

import json
from collections.abc import Generator
from copy import deepcopy
from typing import Any

import pandas as pd
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from palimpzest.core.data import context
from palimpzest.core.lib.schemas import (
    AUDIO_FIELD_TYPES,
    IMAGE_FIELD_TYPES,
    AudioBase64,
    AudioFilepath,
    ImageBase64,
    ImageFilepath,
    ImageURL,
    project,
    union_schemas,
)
from palimpzest.core.models import ExecutionStats, PlanStats, RecordOpStats
from palimpzest.utils.hash_helpers import hash_for_id


class DataRecord:
    """A DataRecord is a single record of data matching some schema defined by a BaseModel."""

    def __init__(
        self,
        data_item: BaseModel,
        source_indices: str | int | list[str | int],
        parent_ids: str | list[str] | None = None,
        cardinality_idx: int | None = None,
    ):
        # check that source_indices are provided
        assert source_indices is not None, "Every DataRecord must be constructed with source index (or indices)"

        # normalize to list[str]
        if not isinstance(source_indices, list):
            source_indices = [source_indices]

        # normalize to list[str]
        if isinstance(parent_ids, str):
            parent_ids = [parent_ids]

        # data for the data record
        self._data_item = data_item

        # the index in the root Dataset from which this DataRecord is derived;
        # each source index takes the form: f"{root_dataset.id}-{idx}"
        self._source_indices = sorted(source_indices)

        # the id(s) of the parent record(s) from which this DataRecord is derived
        self._parent_ids = parent_ids

        # store the cardinality index
        self._cardinality_idx = cardinality_idx

        # indicator variable which may be flipped by filter operations to signal when a record has been filtered out
        self._passed_operator = True

        # NOTE: Record ids are hashed based on:
        # 0. their schema (keys)
        # 1. their parent record id(s) (or source_indices if there is no parent record)
        # 2. their index in the fan out (if this is in a one-to-many operation)
        #
        # We currently do NOT hash just based on record content (i.e. schema (key, value) pairs)
        # because multiple outputs for a given operation may have the exact same
        # schema (key, value) pairs.
        #
        # We may revisit this hashing scheme in the future.

        # unique identifier for the record
        schema_fields = sorted(list(type(data_item).model_fields))
        id_str = (
            str(schema_fields) + str(parent_ids) if parent_ids is not None else str(self._source_indices)
            if cardinality_idx is None
            else str(schema_fields) + str(cardinality_idx) + str(parent_ids) if parent_ids is not None else str(self._source_indices)
        )
        self._id = hash_for_id(id_str)


    # TODO: raise an exception if one of these fields is present in the schema
    # - put these in a constant list up top
    # - import the constant list in Dataset (if possible) and check at plan creation time
    def __setattr__(self, name: str, value: Any, /) -> None:
        if name in ["_data_item", "_source_indices", "_parent_ids", "_cardinality_idx", "_passed_operator", "_id"]:
            super().__setattr__(name, value)
        else:
            setattr(self._data_item, name, value)


    def __getattr__(self, name: str) -> Any:
        return getattr(self._data_item, name)


    def __getitem__(self, field: str) -> Any:
        return getattr(self._data_item, field)


    def __setitem__(self, field: str, value: Any) -> None:
        setattr(self._data_item, field, value)


    def __str__(self, truncate: int | None = 15) -> str:
        if truncate is not None:
            items = (f"{k}={str(v)[:truncate]!r}{'...' if len(str(v)) > truncate else ''}" for k, v in sorted(self._data_item.model_dump().items()))
        else:
            items = (f"{k}={v!r}" for k, v in sorted(self._data_item.model_dump().items()))
        return "{}({})".format(type(self).__name__, ", ".join(items))


    def __repr__(self) -> str:
        return self.__str__(truncate=None)


    def __eq__(self, other):
        return isinstance(other, DataRecord) and self._data_item == other._data_item


    def __hash__(self):
        return hash(self.to_json_str(bytes_to_str=True, sorted=True))


    def __iter__(self):
        yield from self._data_item.__iter__()


    def get_field_names(self):
        return list(type(self._data_item).model_fields.keys())


    def get_field_type(self, field_name: str) -> FieldInfo:
        return type(self._data_item).model_fields[field_name]

    @property
    def schema(self) -> type[BaseModel]:
        return type(self._data_item)

    def copy(self) -> DataRecord:
        # get the set of fields to copy from the parent record
        copy_field_names = [field.split(".")[-1] for field in self.get_field_names()]

        # copy field types and values from the parent
        data_item = {field_name: self[field_name] for field_name in copy_field_names}

        # make copy of the current record
        new_dr = DataRecord(
            self.schema(**data_item),
            source_indices=self._source_indices,
            parent_ids=self._parent_ids,
            cardinality_idx=self._cardinality_idx,
        )

        # copy the passed_operator attribute
        new_dr._passed_operator = self._passed_operator

        return new_dr

    @staticmethod
    def from_parent(
        schema: type[BaseModel],
        data_item: dict,
        parent_record: DataRecord,
        project_cols: list[str] | None = None,
        cardinality_idx: int | None = None,
    ) -> DataRecord:
        # if project_cols is None, then the new schema is a union of the provided schema and parent_record.schema;
        # if project_cols is an empty list, then the new schema is simply the provided schema
        # otherwise, it's a ProjectSchema
        new_schema = None
        if project_cols is None:
            new_schema = union_schemas([schema, parent_record.schema])
        elif project_cols == []:
            new_schema = schema
        else:
            new_schema = union_schemas([schema, parent_record.schema])
            new_schema = project(new_schema, project_cols)

        # get the set of fields and field descriptions to copy from the parent record
        copy_field_names = parent_record.get_field_names() if project_cols is None else project_cols
        copy_field_names = [field.split(".")[-1] for field in copy_field_names]

        # copy fields from the parent
        data_item.update({field_name: parent_record[field_name] for field_name in copy_field_names})

        # corner-case: wrap values in lists if the new schema expects a list but the data item has a single value
        for field_name, field_info in new_schema.model_fields.items():
            field_should_be_list = hasattr(field_info.annotation, '__origin__') and field_info.annotation.__origin__ is list
            field_is_not_list = field_name in data_item and not isinstance(data_item[field_name], list)
            if field_should_be_list and field_is_not_list:
                data_item[field_name] = [data_item[field_name]]

        # make new record which has parent_record as its parent (and the same source_indices)
        new_dr = DataRecord(
            new_schema(**data_item),
            source_indices=parent_record._source_indices,
            parent_ids=[parent_record._id],
            cardinality_idx=cardinality_idx,
        )

        return new_dr

    @staticmethod
    def from_agg_parents(
        data_item: BaseModel,
        parent_records: DataRecordSet,
        cardinality_idx: int | None = None,
    ) -> DataRecord:
        # flatten source indices from all parents
        source_indices = [
            source_idx
            for parent_record in parent_records
            for source_idx in parent_record._source_indices
        ]

        # make new record which has all parent records as its parents
        return DataRecord(
            data_item,
            source_indices=source_indices,
            parent_ids=[parent_record._id for parent_record in parent_records],
            cardinality_idx=cardinality_idx,
        )

    @staticmethod
    def from_join_parents(
        schema: type[BaseModel],
        left_parent_record: DataRecord | None,
        right_parent_record: DataRecord | None,
        project_cols: list[str] | None = None,
        cardinality_idx: int = None,
    ) -> DataRecord:
        # get the set of fields and field descriptions to copy from the parent record(s)
        left_copy_field_names = [] if left_parent_record is None else (
            left_parent_record.get_field_names()
            if project_cols is None
            else [col for col in project_cols if col in left_parent_record.get_field_names()]
        )
        right_copy_field_names = [] if right_parent_record is None else (
            right_parent_record.get_field_names()
            if project_cols is None
            else [col for col in project_cols if col in right_parent_record.get_field_names()]
        )
        left_copy_field_names = [field.split(".")[-1] for field in left_copy_field_names]
        right_copy_field_names = [field.split(".")[-1] for field in right_copy_field_names]

        # copy fields from the parents
        data_item = {field_name: left_parent_record[field_name] for field_name in left_copy_field_names}
        for field_name in right_copy_field_names:
            new_field_name = field_name
            if field_name in left_copy_field_names:
                new_field_name = f"{field_name}_right"
            data_item[new_field_name] = right_parent_record[field_name]

        # for any missing fields in the schema, set them to None
        for field_name in schema.model_fields:
            if field_name not in data_item:
                data_item[field_name] = None

        # make new record which has left and right parent record as its parents
        left_parent_source_indices = [] if left_parent_record is None else list(left_parent_record._source_indices)
        right_parent_source_indices = [] if right_parent_record is None else list(right_parent_record._source_indices)
        left_parent_record_id = [] if left_parent_record is None else [left_parent_record._id]
        right_parent_record_id = [] if right_parent_record is None else [right_parent_record._id]
        new_dr = DataRecord(
            schema(**data_item),
            source_indices=left_parent_source_indices + right_parent_source_indices,
            parent_ids=left_parent_record_id + right_parent_record_id,
            cardinality_idx=cardinality_idx,
        )

        return new_dr

    @staticmethod
    def to_df(records: list[DataRecord], project_cols: list[str] | None = None) -> pd.DataFrame:
        if len(records) == 0:
            return pd.DataFrame()

        fields = records[0].get_field_names()
        if project_cols is not None and len(project_cols) > 0:
            fields = [field for field in fields if field in project_cols]

        # convert Context --> str
        for record in records:
            for k in fields:
                if isinstance(record[k], context.Context):
                    record[k] = record[k].description

        return pd.DataFrame([
            {k: record[k] for k in fields}
            for record in records
        ])

    def to_json_str(self, include_bytes: bool = True, bytes_to_str: bool = False, project_cols: list[str] | None = None, sorted: bool = False):
        """Return a JSON representation of this DataRecord"""
        record_dict = self.to_dict(include_bytes, bytes_to_str, project_cols, sorted)
        return json.dumps(record_dict, indent=2)

    def to_dict(self, include_bytes: bool = True, bytes_to_str: bool = False, project_cols: list[str] | None = None, _sorted: bool = False, mask_filepaths: bool = False):
        """Return a dictionary representation of this DataRecord"""
        # TODO(chjun): In case of numpy types, the json.dumps will fail. Convert to native types.
        # Better ways to handle this.
        field_values = {
            k: v.description if isinstance(v, context.Context) else v
            for k, v in self._data_item.model_dump().items()
        }
        dct = pd.Series(field_values).to_dict()

        if project_cols is not None and len(project_cols) > 0:
            project_field_names = set(field.split(".")[-1] for field in project_cols)
            dct = {k: v for k, v in dct.items() if k in project_field_names}

        if not include_bytes:
            bytes_field_types = [bytes, list[bytes], bytes | None, list[bytes] | None, bytes | Any, list[bytes] | Any]
            bytes_field_types += AUDIO_FIELD_TYPES + IMAGE_FIELD_TYPES
            for k in dct:
                field_type = self.get_field_type(k)
                if field_type.annotation in bytes_field_types:
                    dct[k] = "<bytes>"

        if bytes_to_str:
            for k, v in dct.items():
                if isinstance(v, bytes):
                    dct[k] = v.decode("utf-8")
                elif isinstance(v, list) and len(v) > 0 and any([isinstance(elt, bytes) for elt in v]):
                    dct[k] = [elt.decode("utf-8") if isinstance(elt, bytes) else elt for elt in v]

        if _sorted:
            dct = dict(sorted(dct.items()))

        if mask_filepaths:
            for k in dct:
                field_type = self.get_field_type(k)
                if field_type.annotation in [AudioBase64, AudioFilepath, ImageBase64, ImageFilepath, ImageURL]:
                    dct[k] = "<bytes>"

        return deepcopy(dct)


class DataRecordSet:
    """
    A DataRecordSet contains a list of DataRecords that share the same schema, same parent(s), and same source(s).

    We explicitly check that this is True.

    The record_op_stats could be empty if the DataRecordSet is not from executing an operator.
    """
    def __init__(
            self,
            data_records: list[DataRecord],
            record_op_stats: list[RecordOpStats],
            field_to_score_fn: dict[str, str | callable] | None = None,
            input: int | DataRecord | list[DataRecord] | tuple[list[DataRecord]] | None = None,
        ):
        # set data_records, parent_ids, and source_indices; note that it is possible for
        # data_records to be an empty list in the event of a failed convert
        self.data_records = data_records
        self.parent_ids = data_records[0]._parent_ids if len(data_records) > 0 else None
        self.source_indices = data_records[0]._source_indices if len(data_records) > 0 else None
        self.schema = data_records[0].schema if len(data_records) > 0 else None

        # the input to the operator which produced the data_records; type is tuple[DataRecord] | tuple[int]
        # - for scan operators, input is a singleton tuple[int] which wraps the source_idx, e.g.: (source_idx,)
        # - for join operators, input is a tuple with one entry for the left input DataRecord and one entry for the right input DataRecord
        # - for aggregate operators, input is a tuple with all the input DataRecords to the aggregation
        # - for all other operaotrs, input is a singleton tuple[DataRecord] which wraps the single input
        self.input = input

        # set statistics for generating these records
        self.record_op_stats = record_op_stats

        # assign field_to_score_fn if provided
        self.field_to_score_fn = {} if field_to_score_fn is None else field_to_score_fn

    def get_total_cost(self) -> float:
        return sum([record_op_stats.cost_per_record for record_op_stats in self.record_op_stats])

    def get_field_to_score_fn(self) -> dict[str, str | callable]:
        return self.field_to_score_fn

    def __getitem__(self, slice) -> DataRecord | list[DataRecord]:
        return self.data_records[slice]

    def __len__(self) -> int:
        return len(self.data_records)

    def __iter__(self) -> Generator[DataRecord]:
        yield from self.data_records


class DataRecordCollection:
    """
    A DataRecordCollection contains a list of DataRecords.

    This is a wrapper class for list[DataRecord] to support more advanced features for output of execute().

    The difference between DataRecordSet and DataRecordCollection 

    Goal: 
        DataRecordSet is a set of DataRecords that share the same schema, same parents, and same sources.
        DataRecordCollection is a general wrapper for list[DataRecord].
    
    Usage:
        DataRecordSet is used for the output of executing an operator.
        DataRecordCollection is used for the output of executing a query, we definitely could extend it to support more advanced features for output of execute().
    """
    def __init__(self, data_records: list[DataRecord], execution_stats: ExecutionStats | None = None, plan_stats: PlanStats | None = None):
        self.data_records = data_records
        self.execution_stats = execution_stats
        self.plan_stats = plan_stats
        self.executed_plans = self._get_executed_plans()

    def __iter__(self) -> Generator[DataRecord]:
        """Allow iterating directly over the data records"""
        yield from self.data_records

    def __len__(self):
        """Return the number of records in the collection"""
        return len(self.data_records)

    def to_df(self, cols: list[str] | None = None):
        return DataRecord.to_df(self.data_records, cols)

    def _get_executed_plans(self):
        if self.plan_stats is not None:
            return [self.plan_stats.plan_str]
        elif self.execution_stats is not None:
            return list(self.execution_stats.plan_strs.values())
        else:
            return None



import logging

from palimpzest.core.data.dataset import Dataset
from palimpzest.core.elements.records import DataRecord, DataRecordCollection
from palimpzest.core.models import ExecutionStats, PlanStats
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_strategy import ExecutionStrategy, SentinelExecutionStrategy
from palimpzest.query.optimizer.cost_model import SampleBasedCostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.utils.hash_helpers import hash_for_id
from palimpzest.validator.validator import Validator

logger = logging.getLogger(__name__)

class QueryProcessor:
    """
    Processes queries through the complete pipeline:
    1. Optimization phase: Plan generation and selection
    2. Execution phase: Plan execution and result collection
    3. Result phase: Statistics gathering and result formatting
    """
    def __init__(
        self,
        dataset: Dataset,
        optimizer: Optimizer,
        execution_strategy: ExecutionStrategy,
        sentinel_execution_strategy: SentinelExecutionStrategy | None,
        num_samples: int | None = None,
        train_dataset: dict[str, Dataset] | None = None,
        validator: Validator | None = None,
        scan_start_idx: int = 0,
        verbose: bool = False,
        progress: bool = True,
        max_workers: int | None = None,
        policy: Policy | None = None,
        available_models: list[str] | None = None,
        **kwargs,  # needed in order to provide compatibility with QueryProcessorConfig
    ):
        """
        Initialize QueryProcessor with optional custom components.
        
        Args:
            dataset: Dataset to process
            TODO
        """
        self.dataset = dataset
        self.optimizer = optimizer
        self.execution_strategy = execution_strategy
        self.sentinel_execution_strategy = sentinel_execution_strategy
        self.num_samples = num_samples
        self.train_dataset = train_dataset
        self.validator = validator
        self.scan_start_idx = scan_start_idx
        self.verbose = verbose
        self.progress = progress
        self.max_workers = max_workers
        self.policy = policy
        self.available_models = available_models

        if self.verbose:
            print("Available models: ", self.available_models)

        logger.info(f"Initialized QueryProcessor {self.__class__.__name__}")
        logger.debug(f"QueryProcessor initialized with config: {self.__dict__}")

    def execution_id(self) -> str:
        """
        Hash of the class parameters.
        """
        id_str = ""
        for attr, value in self.__dict__.items():
            if not attr.startswith("_"):
                id_str += f"{attr}={value},"

        return hash_for_id(id_str)

    def _create_sentinel_plan(self, train_dataset: dict[str, Dataset] | None) -> SentinelPlan:
        """
        Generates and returns a SentinelPlan for the given dataset.
        """
        # create a new optimizer and update its strategy to SENTINEL
        optimizer = self.optimizer.deepcopy_clean()
        optimizer.update_strategy(OptimizationStrategyType.SENTINEL)

        # create copy of dataset, but change its root Dataset(s) to the validation Dataset(s)
        dataset = self.dataset.copy()
        if train_dataset is not None:
            dataset._set_root_datasets(train_dataset)
            dataset._generate_unique_logical_op_ids()

        # get the sentinel plan for the given dataset
        sentinel_plans = optimizer.optimize(dataset)
        sentinel_plan = sentinel_plans[0]

        return sentinel_plan

    def _execute_best_plan(self, dataset: Dataset, optimizer: Optimizer) -> tuple[list[DataRecord], list[PlanStats]]:
        # get the optimal plan according to the optimizer
        plans = optimizer.optimize(dataset)
        final_plan = plans[0]

        # execute the plan
        records, plan_stats = self.execution_strategy.execute_plan(plan=final_plan)

        # return the output records and plan stats
        return records, [plan_stats]

    def execute(self) -> DataRecordCollection:
        logger.info(f"Executing {self.__class__.__name__}")

        # create execution stats
        execution_stats = ExecutionStats(execution_id=self.execution_id())
        execution_stats.start()

        # if the user provides a validator, we perform optimization
        if self.validator is not None:
            # create sentinel plan
            sentinel_plan = self._create_sentinel_plan(self.train_dataset)

            # generate sample execution data
            if self.train_dataset is not None:
                sentinel_plan_stats = self.sentinel_execution_strategy.execute_sentinel_plan(sentinel_plan, self.train_dataset, self.validator)

            else:
                train_dataset = self.dataset._get_root_datasets()
                sentinel_plan_stats = self.sentinel_execution_strategy.execute_sentinel_plan(sentinel_plan, train_dataset, self.validator)

            # update the execution stats to account for the work done in optimization
            execution_stats.add_plan_stats(sentinel_plan_stats)
            execution_stats.finish_optimization()

            # (re-)initialize the optimizer
            self.optimizer = self.optimizer.deepcopy_clean()

            # construct the CostModel with any sample execution data we've gathered
            cost_model = SampleBasedCostModel(sentinel_plan_stats, self.verbose)
            self.optimizer.update_cost_model(cost_model)

        # execute plan(s) according to the optimization strategy
        records, plan_stats = self._execute_best_plan(self.dataset, self.optimizer)

        # update the execution stats to account for the work to execute the final plan
        execution_stats.add_plan_stats(plan_stats)
        execution_stats.finish()

        # construct and return the DataRecordCollection
        result = DataRecordCollection(records, execution_stats=execution_stats)
        logger.info(f"Done executing {self.__class__.__name__}")

        return result



"""This file contains prompts for filter operations."""

### BASE PROMPTS ###
FILTER_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Output TRUE if the context satisfies the filter condition, and FALSE otherwise.

Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

CONTEXT:
{{{example_context}}}{image_disclaimer}{audio_disclaimer}

FILTER CONDITION: {example_filter_condition}

Let's think step-by-step in order to answer the question.

REASONING: {example_reasoning}

ANSWER: TRUE
---
"""

FILTER_NO_REASONING_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Output TRUE if the context satisfies the filter condition, and FALSE otherwise.

Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

CONTEXT:
{{{example_context}}}{image_disclaimer}{audio_disclaimer}

FILTER CONDITION: {example_filter_condition}

ANSWER: TRUE
---
"""

FILTER_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Output TRUE if the context satisfies the filter condition, and FALSE otherwise.
{desc_section}
Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---
---
INPUT FIELDS:
{input_fields_desc}

CONTEXT:
{context}<<image-audio-placeholder>>

FILTER CONDITION: {filter_condition}

Let's think step-by-step in order to answer the question.

REASONING: """

FILTER_NO_REASONING_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Output TRUE if the context satisfies the filter condition, and FALSE otherwise.
{desc_section}
Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---
---
INPUT FIELDS:
{input_fields_desc}

CONTEXT:
{context}<<image-audio-placeholder>>

FILTER CONDITION: {filter_condition}

ANSWER: """



from __future__ import annotations

import os
import pickle

import chromadb
import chromadb.utils.embedding_functions as embedding_functions
import tiktoken

from palimpzest.constants import PZ_DIR
from palimpzest.core.data import context


class ContextNotFoundError(Exception):
    pass


class ContextManager:
    """
    This class manages the long-term storage of `Contexts`. Each new `Context` is added to
    the `ContextManager` and serialized to disk. `Contexts` are also indexed, which enables
    PZ to search for `Context(s)` which may support `search()` and `compute()` operations.
    """
    def __init__(self):
        # create directory with serialized contexts (if it doesn't already exist)
        self.context_dir = os.path.join(PZ_DIR, "contexts")
        os.makedirs(self.context_dir, exist_ok=True)

        # create vector store (if it doesn't already exist)
        self.chroma_dir = os.path.join(PZ_DIR, "chroma")
        os.makedirs(self.chroma_dir, exist_ok=True)
        self.chroma_client = chromadb.PersistentClient(self.chroma_dir)

        # pick embedding function based on presence of API key(s)
        self.emb_fn = None
        if os.getenv("OPENAI_API_KEY"):
            self.emb_fn = embedding_functions.OpenAIEmbeddingFunction(
                api_key=os.getenv("OPENAI_API_KEY"),
                model_name="text-embedding-3-small"
            )

        self.index = self.chroma_client.get_or_create_collection("contexts", embedding_function=self.emb_fn)

    @staticmethod
    def from_pkl(path: str) -> context.Context:
        """Load a `Context` from its serialized pickle file."""
        with open(path, "rb") as f:
            context = pickle.load(f)

        return context

    @staticmethod
    def to_pkl(context: context.Context, path: str) -> None:
        """Write the given `Context` to a pickle file at the provided `path`."""
        with open(path, "wb") as f:
            pickle.dump(context, f)

    def num_tokens_from_string(self, string: str, encoding_name: str) -> int:
        """Returns the number of tokens in a text string."""
        encoding = tiktoken.get_encoding(encoding_name)
        num_tokens = len(encoding.encode(string))
        return num_tokens

    def add_context(self, context: context.Context, update: bool = False) -> None:
        """
        Add the new `Context` to the `ContextManager` by serializing and writing it to disk.

        Args:
            context (`Context`): the context to add to the `ContextManager`
            update (`bool`): whether or not to update an existing context

        TODO: track cost
        """
        # return early if the context already exists and we're not performing an update
        id = context.id
        context_path = os.path.join(self.context_dir, f"{id}.pkl")
        if os.path.exists(context_path) and update is False:
            return

        # write the context to disk
        ContextManager.to_pkl(context, context_path)

        # compute number of tokens in context.description
        description = context.description
        while self.num_tokens_from_string(description, "cl100k_base") > 8192:
            description = description[:int(0.9*len(description))]
 
        # add context to vector store
        context_embeddings = self.emb_fn([description])
        context_payload = {
            "ids": [context.id],
            "embeddings": context_embeddings,
            "metadatas": [{"id": context.id, "materialized": context.materialized}],
            "documents": [context.description],
        }
        if update:
            self.index.update(**context_payload)
        else:
            self.index.add(**context_payload)

    def update_context(self, id: str, description: str, materialized: bool = True) -> None:
        """
        Update an existing `Context` with the given `id` to have the given `description`.
        
        Args:
            id (str): the id of the updated `Context`
            description (str): the update to the description for the specified `Context`
            materialized (bool): boolean to set the materialization status of the `Context`

        Raises:
            ContextNotFoundError: if the given `id` doesn't point to a `Context` in the `ContextManger`.
        """
        context = self.get_context(id)
        new_description = context.description + description  # TODO: should description have RESULT replaced on update? as opposed to appending? should description be some pydantic BaseModel?
        context.set_description(new_description)
        context.set_materialized(materialized)
        self.add_context(context, update=True)

    def get_context(self, id: str) -> context.Context:
        """
        Returns the `Context` specified by the given `id`.

        Args:
            id (str): the id of the retrieved `Context`

        Returns:
            `Context`: the specified `Context`.
        """
        context_path = os.path.join(self.context_dir, f"{id}.pkl")
        try:
            return ContextManager.from_pkl(context_path)
        except FileNotFoundError as err:
            raise ContextNotFoundError from err

    def search_context(self, query: str, k: int = 1, where: dict | None = None) -> list[context.Context]:
        """
        Returns the top-k most relevant `Context(s)` for the given query. If provided,
        the where dictionary will be used to filter the search results.

        TODO:
        3) update CostModel to account for benefit of using existing Context(s)
        ---
        4) unit test
        5) track cost
        """
        # embed the search query
        query_embeddings = self.emb_fn([query])

        # look up ids of most similar contexts
        results = self.index.query(
            query_embeddings=query_embeddings,
            n_results=k,
            where=where,
        )
        ids = results["ids"][0]

        # load and return Context objects
        contexts = []
        for id in ids:
            context_path = os.path.join(self.context_dir, f"{id}.pkl")
            contexts.append(ContextManager.from_pkl(context_path))

        return contexts



from __future__ import annotations

import contextlib
import time
from typing import Any

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_NUM_GROUPS,
    NAIVE_EST_NUM_INPUT_TOKENS,
    NAIVE_EST_NUM_OUTPUT_TOKENS,
    AggFunc,
    Model,
    PromptStrategy,
)
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import Average, Count, Max, Min, Sum
from palimpzest.core.models import OperatorCostEstimates, RecordOpStats
from palimpzest.query.generators.generators import Generator
from palimpzest.query.operators.physical import PhysicalOperator


class AggregateOp(PhysicalOperator):
    """
    Aggregate operators accept a list of candidate DataRecords as input to their
    __call__ methods. Thus, we use a slightly modified abstract base class for
    these operators.
    """
    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        raise NotImplementedError("Using __call__ from abstract method")


class ApplyGroupByOp(AggregateOp):
    """
    Implementation of a GroupBy operator. This operator groups records by a set of fields
    and applies a function to each group. The group_by_sig object contains the fields to
    group by and the aggregation functions to apply to each group.
    """
    def __init__(self, group_by_sig: GroupBySig, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.group_by_sig = group_by_sig

    def __str__(self):
        op = super().__str__()
        op += f"    Group-by Signature: {str(self.group_by_sig)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"group_by_sig": str(self.group_by_sig.serialize()), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"group_by_sig": self.group_by_sig, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the groupby takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=NAIVE_EST_NUM_GROUPS,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    @staticmethod
    def agg_init(func):
        if func.lower() == "count":
            return 0
        elif func.lower() == "average":
            return (0, 0)
        elif func.lower() == "sum":
            return 0
        elif func.lower() == "min":
            return float("inf")
        elif func.lower() == "max":
            return float("-inf")
        elif func.lower() == "list":
            return []
        elif func.lower() == "set":
            return set()
        else:
            raise Exception("Unknown agg function " + func)

    @staticmethod
    def agg_merge(func, state, val):
        if func.lower() == "count":
            return state + 1
        elif func.lower() == "average":
            sum_, cnt = state
            if val is None:
                return (sum_, cnt)
            return (sum_ + val, cnt + 1)
        elif func.lower() == "sum":
            if val is None:
                return state
            return state + sum(val) if isinstance(val, list) else state + val
        elif func.lower() == "min":
            if val is None:
                return state
            return min(state, min(val) if isinstance(val, list) else val)
        elif func.lower() == "max":
            if val is None:
                return state
            return max(state, max(val) if isinstance(val, list) else val)
        elif func.lower() == "list":
            state.append(val)
            return state
        elif func.lower() == "set":
            state.add(val)
            return state
        else:
            raise Exception("Unknown agg function " + func)

    @staticmethod
    def agg_final(func, state):
        if func.lower() in ["count", "sum", "min", "max", "list", "set"]:
            return state
        elif func.lower() == "average":
            sum, cnt = state
            return float(sum) / cnt if cnt > 0 else None
        else:
            raise Exception("Unknown agg function " + func)

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # build group array
        agg_state = {}
        for candidate in candidates:
            group = ()
            for f in self.group_by_sig.group_by_fields:
                if not hasattr(candidate, f):
                    raise TypeError(f"ApplyGroupByOp record missing expected field {f}")
                group = group + (getattr(candidate, f),)
            if group in agg_state:
                state = agg_state[group]
            else:
                state = []
                for fun in self.group_by_sig.agg_funcs:
                    state.append(ApplyGroupByOp.agg_init(fun))
            for i in range(0, len(self.group_by_sig.agg_funcs)):
                fun = self.group_by_sig.agg_funcs[i]
                if not hasattr(candidate, self.group_by_sig.agg_fields[i]):
                    raise TypeError(f"ApplyGroupByOp record missing expected field {self.group_by_sig.agg_fields[i]}")
                field = getattr(candidate, self.group_by_sig.agg_fields[i])
                state[i] = ApplyGroupByOp.agg_merge(fun, state[i], field)
            agg_state[group] = state

        # return list of data records (one per group)
        drs: list[DataRecord] = []
        group_by_fields = self.group_by_sig.group_by_fields
        agg_fields = self.group_by_sig.get_agg_field_names()
        for g in agg_state:
            # build up data item
            data_item = {}
            for i in range(0, len(g)):
                k = g[i]
                data_item[group_by_fields[i]] = k
            vals = agg_state[g]
            for i in range(0, len(vals)):
                v = ApplyGroupByOp.agg_final(self.group_by_sig.agg_funcs[i], vals[i])
                data_item[agg_fields[i]] = v

            # create new DataRecord
            schema = self.group_by_sig.output_schema()
            data_item = schema(**data_item)
            dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)
            drs.append(dr)

        # create RecordOpStats objects
        total_time = time.time() - start_time
        record_op_stats_lst = []
        for dr in drs:
            record_op_stats = RecordOpStats(
                record_id=dr._id,
                record_parent_ids=dr._parent_ids,
                record_source_indices=dr._source_indices,
                record_state=dr.to_dict(include_bytes=False),
                full_op_id=self.get_full_op_id(),
                logical_op_id=self.logical_op_id,
                op_name=self.op_name(),
                time_per_record=total_time / len(drs),
                cost_per_record=0.0,
                op_details={k: str(v) for k, v in self.get_id_params().items()},
            )
            record_op_stats_lst.append(record_op_stats)

        # construct and return DataRecordSet
        return DataRecordSet(drs, record_op_stats_lst)


class AverageAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        # enforce that output schema is correct
        assert kwargs["output_schema"].model_fields.keys() == Average.model_fields.keys(), "AverageAggregateOp requires output_schema to be Average"

        # enforce that input schema is a single numeric field
        input_field_types = list(kwargs["input_schema"].model_fields.values())
        assert len(input_field_types) == 1, "AverageAggregateOp requires input_schema to have exactly one field"
        numeric_field_types = [
            bool, int, float, int | float,
            bool | None, int | None, float | None, int | float | None,
            bool | Any, int | Any, float | Any, int | float | Any,
            bool | None | Any, int | None | Any, float | None | Any, int | float | None | Any,
        ]
        is_numeric = input_field_types[0].annotation in numeric_field_types
        assert is_numeric, f"AverageAggregateOp requires input_schema to have a numeric field type, i.e. one of: {numeric_field_types}\nGot: {input_field_types[0]}"

        # call parent constructor
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # NOTE: we currently do not guarantee that input values conform to their specified type;
        #       as a result, we simply omit any values which do not parse to a float from the average
        # NOTE: right now we perform a check in the constructor which enforces that the input_schema
        #       has a single field which is numeric in nature; in the future we may want to have a
        #       cleaner way of computing the value (rather than `float(list(candidate...))` below)
        summation, total = 0, 0
        for candidate in candidates:
            try:
                summation += float(list(candidate.to_dict().values())[0])
                total += 1
            except Exception:
                pass
        data_item = Average(average=summation / total)
        dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
        )

        return DataRecordSet([dr], [record_op_stats])


class SumAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        # enforce that output schema is correct
        assert kwargs["output_schema"].model_fields.keys() == Sum.model_fields.keys(), "SumAggregateOp requires output_schema to be Sum"

        # enforce that input schema is a single numeric field
        input_field_types = list(kwargs["input_schema"].model_fields.values())
        assert len(input_field_types) == 1, "SumAggregateOp requires input_schema to have exactly one field"
        numeric_field_types = [
            bool, int, float, int | float,
            bool | None, int | None, float | None, int | float | None,
            bool | Any, int | Any, float | Any, int | float | Any,
            bool | None | Any, int | None | Any, float | None | Any, int | float | None | Any,
        ]
        is_numeric = input_field_types[0].annotation in numeric_field_types
        assert is_numeric, f"SumAggregateOp requires input_schema to have a numeric field type, i.e. one of: {numeric_field_types}\nGot: {input_field_types[0]}"

        # call parent constructor
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # NOTE: we currently do not guarantee that input values conform to their specified type;
        #       as a result, we simply omit any values which do not parse to a float from the average
        # NOTE: right now we perform a check in the constructor which enforces that the input_schema
        #       has a single field which is numeric in nature; in the future we may want to have a
        #       cleaner way of computing the value (rather than `float(list(candidate...))` below)
        summation = 0
        for candidate in candidates:
            with contextlib.suppress(Exception):
                summation += float(list(candidate.to_dict().values())[0])
        data_item = Sum(sum=summation)
        dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
        )

        return DataRecordSet([dr], [record_op_stats])


class CountAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        # enforce that output schema is correct
        assert kwargs["output_schema"].model_fields.keys() == Count.model_fields.keys(), "CountAggregateOp requires output_schema to be Count"

        # call parent constructor
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # create new DataRecord
        data_item = Count(count=len(candidates))
        dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])


class MinAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        # enforce that output schema is correct
        assert kwargs["output_schema"].model_fields.keys() == Min.model_fields.keys(), "MinAggregateOp requires output_schema to be Min"

        # call parent constructor
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # create new DataRecord
        min = float("inf")
        for candidate in candidates:
            try:  # noqa: SIM105
                min = min(float(list(candidate.to_dict().values())[0]), min)
            except Exception:
                pass
        data_item = Min(min=min if min != float("inf") else None)
        dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_ids=dr.parent_ids,
            record_source_indices=dr.source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])


class MaxAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        # enforce that output schema is correct
        assert kwargs["output_schema"].model_fields.keys() == Max.model_fields.keys(), "MaxAggregateOp requires output_schema to be Max"

        # call parent constructor
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # create new DataRecord
        
        max = float("-inf")
        for candidate in candidates:
            try:  # noqa: SIM105
                max = max(float(list(candidate.to_dict().values())[0]), max)
            except Exception:
                pass
        data_item = Max(max=max if max != float("-inf") else None)
        dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_ids=dr.parent_ids,
            record_source_indices=dr.source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])


class SemanticAggregate(AggregateOp):

    def __init__(self, agg_str: str, model: Model, prompt_strategy: PromptStrategy = PromptStrategy.AGG, reasoning_effort: str | None = None, *args, **kwargs):
        # call parent constructor
        super().__init__(*args, **kwargs)
        self.agg_str = agg_str
        self.model = model
        self.prompt_strategy = prompt_strategy
        self.reasoning_effort = reasoning_effort
        if model is not None:
            self.generator = Generator(model, prompt_strategy, reasoning_effort, self.api_base)

    def __str__(self):
        op = super().__str__()
        op += f"    Prompt Strategy: {self.prompt_strategy}\n"
        op += f"    Reasoning Effort: {self.reasoning_effort}\n"
        op += f"    Agg: {str(self.agg_str)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "agg_str": self.agg_str,
            "model": None if self.model is None else self.model.value,
            "prompt_strategy": None if self.prompt_strategy is None else self.prompt_strategy.value,
            "reasoning_effort": self.reasoning_effort,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "agg_str": self.agg_str,
            "model": self.model,
            "prompt_strategy": self.prompt_strategy,
            "reasoning_effort": self.reasoning_effort,
            **op_params,
        }

        return op_params

    def get_model_name(self) -> str:
        return self.model.value

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Compute naive cost estimates for the LLMConvert operation. Implicitly, these estimates
        assume the use of a single LLM call for each input record. Child classes of LLMConvert
        may call this function through super() and adjust these estimates as needed (or they can
        completely override this function).
        """
        # estimate number of input and output tokens from source
        est_num_input_tokens = NAIVE_EST_NUM_INPUT_TOKENS * source_op_cost_estimates.cardinality
        est_num_output_tokens = NAIVE_EST_NUM_OUTPUT_TOKENS

        # get est. of conversion time per record from model card;
        model_name = self.model.value
        model_conversion_time_per_record = MODEL_CARDS[model_name]["seconds_per_output_token"] * est_num_output_tokens

        # get est. of conversion cost (in USD) per record from model card
        usd_per_input_token = MODEL_CARDS[model_name].get("usd_per_input_token")
        if getattr(self, "prompt_strategy", None) is not None and self.prompt_strategy.is_audio_prompt():
            usd_per_input_token = MODEL_CARDS[model_name]["usd_per_audio_input_token"]

        model_conversion_usd_per_record = (
            usd_per_input_token * est_num_input_tokens
            + MODEL_CARDS[model_name]["usd_per_output_token"] * est_num_output_tokens
        )

        # estimate quality of output based on the strength of the model being used
        quality = (MODEL_CARDS[model_name]["overall"] / 100.0)

        return OperatorCostEstimates(
            cardinality=1.0,
            time_per_record=model_conversion_time_per_record,
            cost_per_record=model_conversion_usd_per_record,
            quality=quality,
        )

    def __call__(self, candidates: list[DataRecord]) -> DataRecordSet:
        start_time = time.time()

        # TODO: if candidates is an empty list, return an empty DataRecordSet
        if len(candidates) == 0:
            return DataRecordSet([], [])

        # get the set of input fields to use for the operation
        input_fields = self.get_input_fields()

        # get the set of output fields to use for the operation
        fields_to_generate = self.get_fields_to_generate(candidates[0])
        fields = {field: field_type for field, field_type in self.output_schema.model_fields.items() if field in fields_to_generate}

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "output_schema": self.output_schema, "agg_instruction": self.agg_str}

        # generate outputs for all fields in a single query
        field_answers, _, generation_stats, _ = self.generator(candidates, fields, **gen_kwargs)
        assert all([field in field_answers for field in fields]), "Not all fields were generated!"

        # construct data record for the output
        field, value = fields_to_generate[0], field_answers[fields_to_generate[0]][0]
        data_item = self.output_schema(**{field: value})
        dr = DataRecord.from_agg_parents(data_item, parent_records=candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=generation_stats.cost_per_record,
            model_name=self.get_model_name(),
            answer={field: value},
            input_fields=input_fields,
            generated_fields=fields_to_generate,
            total_input_tokens=generation_stats.total_input_tokens,
            total_output_tokens=generation_stats.total_output_tokens,
            total_input_cost=generation_stats.total_input_cost,
            total_output_cost=generation_stats.total_output_cost,
            llm_call_duration_secs=generation_stats.llm_call_duration_secs,
            fn_call_duration_secs=generation_stats.fn_call_duration_secs,
            total_llm_calls=generation_stats.total_llm_calls,
            total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
            image_operation=self.is_image_op(),
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])



CODE_AGENT_SYSTEM_PROMPT = """You are an expert assistant who can solve any task using code blobs. You will be given a task to solve as best you can.
To do so, you have been given access to a list of tools: these tools are basically Python functions which you can call with code.
To solve the task, you must plan forward to proceed in a series of steps, in a cycle of 'Thought:', '<code>', and 'Observation:' sequences.

At each step, in the 'Thought:' sequence, you should first explain your reasoning towards solving the task and the tools that you want to use.
Then in the '<code>' sequence, you should write the code in simple Python. The code sequence must end with '</code>' sequence.
During each intermediate step, you can use 'print()' to save whatever important information you will then need.
These print outputs will then appear in the 'Observation:' field, which will be available as input for the next step.
In the end you have to return a final answer using the `final_answer` tool.

Here are a few examples using notional tools:
---
Task: "Generate an image of the oldest person in this document."

Thought: I will proceed step by step and use the following tools: `document_qa` to find the oldest person in the document, then `image_generator` to generate an image according to the answer.
<code>
answer = document_qa(document=document, question="Who is the oldest person mentioned?")
print(answer)
</code>
Observation: "The oldest person in the document is John Doe, a 55 year old lumberjack living in Newfoundland."

Thought: I will now generate an image showcasing the oldest person.
<code>
image = image_generator("A portrait of John Doe, a 55-year-old man living in Canada.")
final_answer(image)
</code>

---
Task: "What is the result of the following operation: 5 + 3 + 1294.678?"

Thought: I will use python code to compute the result of the operation and then return the final answer using the `final_answer` tool
<code>
result = 5 + 3 + 1294.678
final_answer(result)
</code>

---
Task:
"Answer the question in the variable `question` about the image stored in the variable `image`. The question is in French.
You have been provided with these additional arguments, that you can access using the keys as variables in your python code:
{'question': 'Quel est l'animal sur l'image?', 'image': 'path/to/image.jpg'}"

Thought: I will use the following tools: `translator` to translate the question into English and then `image_qa` to answer the question on the input image.
<code>
translated_question = translator(question=question, src_lang="French", tgt_lang="English")
print(f"The translated question is {translated_question}.")
answer = image_qa(image=image, question=translated_question)
final_answer(f"The answer is {answer}")
</code>

---
Task:
In a 1979 interview, Stanislaus Ulam discusses with Martin Sherwin about other great physicists of his time, including Oppenheimer.
What does he say was the consequence of Einstein learning too much math on his creativity, in one word?

Thought: I need to find and read the 1979 interview of Stanislaus Ulam with Martin Sherwin.
<code>
pages = web_search(query="1979 interview Stanislaus Ulam Martin Sherwin physicists Einstein")
print(pages)
</code>
Observation:
No result found for query "1979 interview Stanislaus Ulam Martin Sherwin physicists Einstein".

Thought: The query was maybe too restrictive and did not find any results. Let's try again with a broader query.
<code>
pages = web_search(query="1979 interview Stanislaus Ulam")
print(pages)
</code>
Observation:
Found 6 pages:
[Stanislaus Ulam 1979 interview](https://ahf.nuclearmuseum.org/voices/oral-histories/stanislaus-ulams-interview-1979/)

[Ulam discusses Manhattan Project](https://ahf.nuclearmuseum.org/manhattan-project/ulam-manhattan-project/)

(truncated)

Thought: I will read the first 2 pages to know more.
<code>
for url in ["https://ahf.nuclearmuseum.org/voices/oral-histories/stanislaus-ulams-interview-1979/", "https://ahf.nuclearmuseum.org/manhattan-project/ulam-manhattan-project/"]:
    whole_page = visit_webpage(url)
    print(whole_page)
    print("\n" + "="*80 + "\n")  # Print separator between pages
</code>
Observation:
Manhattan Project Locations:
Los Alamos, NM
Stanislaus Ulam was a Polish-American mathematician. He worked on the Manhattan Project at Los Alamos and later helped design the hydrogen bomb. In this interview, he discusses his work at
(truncated)

Thought: I now have the final answer: from the webpages visited, Stanislaus Ulam says of Einstein: "He learned too much mathematics and sort of diminished, it seems to me personally, it seems to me his purely physics creativity." Let's answer in one word.
<code>
final_answer("diminished")
</code>

---
Task: "Which city has the highest population: Guangzhou or Shanghai?"

Thought: I need to get the populations for both cities and compare them: I will use the tool `web_search` to get the population of both cities.
<code>
for city in ["Guangzhou", "Shanghai"]:
    print(f"Population {city}:", web_search(f"{city} population")
</code>
Observation:
Population Guangzhou: ['Guangzhou has a population of 15 million inhabitants as of 2021.']
Population Shanghai: '26 million (2019)'

Thought: Now I know that Shanghai has the highest population.
<code>
final_answer("Shanghai")
</code>

---
Task: "What is the current age of the pope, raised to the power 0.36?"

Thought: I will use the tool `wikipedia_search` to get the age of the pope, and confirm that with a web search.
<code>
pope_age_wiki = wikipedia_search(query="current pope age")
print("Pope age as per wikipedia:", pope_age_wiki)
pope_age_search = web_search(query="current pope age")
print("Pope age as per google search:", pope_age_search)
</code>
Observation:
Pope age: "The pope Francis is currently 88 years old."

Thought: I know that the pope is 88 years old. Let's compute the result using python code.
<code>
pope_current_age = 88 ** 0.36
final_answer(pope_current_age)
</code>

Above example were using notional tools that might not exist for you. On top of performing computations in the Python code snippets that you create, you only have access to these tools, behaving like regular python functions:
```python
{%- for tool in tools.values() %}
def {{ tool.name }}({% for arg_name, arg_info in tool.inputs.items() %}{{ arg_name }}: {{ arg_info.type }}{% if not loop.last %}, {% endif %}{% endfor %}) -> {{tool.output_type}}:
    \"\"\"{{ tool.description }}

    Args:
    {%- for arg_name, arg_info in tool.inputs.items() %}
        {{ arg_name }}: {{ arg_info.description }}
    {%- endfor %}
    \"\"\"
{% endfor %}
```

{%- if managed_agents and managed_agents.values() | list %}
You can also give tasks to team members.
Calling a team member works similarly to calling a tool: provide the task description as the 'task' argument. Since this team member is a real human, be as detailed and verbose as necessary in your task description.
You can also include any relevant variables or context using the 'additional_args' argument.
Here is a list of the team members that you can call:
```python
{%- for agent in managed_agents.values() %}
def {{ agent.name }}(task: str, additional_args: dict[str, Any]) -> str:
    \"\"\"{{ agent.description }}

    Args:
        task: Long detailed description of the task.
        additional_args: Dictionary of extra inputs to pass to the managed agent, e.g. images, dataframes, or any other contextual data it may need.
    \"\"\"
{% endfor %}
```
{%- endif %}

Here are the rules you should always follow to solve your task:
1. Always provide a 'Thought:' sequence, and a '<code>' sequence ending with '</code>', else you will fail.
2. Use only variables that you have defined!
3. Always use the right arguments for the tools. DO NOT pass the arguments as a dict as in 'answer = wikipedia_search({'query': "What is the place where James Bond lives?"})', but use the arguments directly as in 'answer = wikipedia_search(query="What is the place where James Bond lives?")'.
4. Take care to not chain too many sequential tool calls in the same code block, especially when the output format is unpredictable. For instance, a call to wikipedia_search has an unpredictable return format, so do not have another tool call that depends on its output in the same block: rather output results with print() to use them in the next block.
5. Call a tool only when needed, and never re-do a tool call that you previously did with the exact same parameters.
6. Don't name any new variable with the same name as a tool: for instance don't name a variable 'final_answer'.
7. Never create any notional variables in our code, as having these in your logs will derail you from the true variables.
8. You can use imports in your code, but only from the following list of modules: {{authorized_imports}}
9. The state persists between code executions: so if in one step you've created variables or imported modules, these will all persist.
10. Don't give up! You're in charge of solving the task, not providing directions to solve it.

{%- if custom_instructions %}
{{custom_instructions}}
{%- endif %}

Now Begin!"""


DATA_DISCOVERY_AGENT_INITIAL_PLAN_PROMPT = """You are a world expert at analyzing a situation to derive facts, and plan accordingly towards solving a task.
Below I will present you a task. You will need to 1. build a survey of facts known or needed to solve the task, then 2. make a plan of action to solve the task.

## 1. Facts survey
You will build a comprehensive preparatory survey of which facts we have at our disposal and which ones we still need.
These "facts" will typically be specific names, dates, values, etc. Be sure to report the facts you look up and derive, as well as the sources where you find these facts.
Your answer should use the below headings:
### 1.1. Facts given in the task
List here the specific facts given in the task that could help you (there might be nothing here).

### 1.2. Facts to look up
List here any facts that we may need to look up.
Also list where to find each of these, for instance a website, a file... - maybe the task contains some sources that you should re-use here.

### 1.3. Facts to derive
List here anything that we want to derive from the above by logical reasoning, for instance computation or simulation.

### 1.4. Fact Sources
List the source of each fact that you have looked up or derived. The source may be a file, a database table, a web page, etc.
Be sure that someone can use these sources to reproduce your work. 

Don't make any assumptions. For each item, provide a thorough reasoning. Do not add anything else on top of the four headings above.

## 2. Plan
Then for the given task, develop a step-by-step high-level plan taking into account the above inputs and list of facts.
This plan should involve individual tasks based on the available tools, that if executed correctly will yield the correct answer.
Do not skip steps, do not add any superfluous steps. Only write the high-level plan, DO NOT DETAIL INDIVIDUAL TOOL CALLS.
After writing the final step of the plan, write the '<end_plan>' tag and stop there.

You can leverage these tools, behaving like regular python functions:
```python
{%- for tool in tools.values() %}
def {{ tool.name }}({% for arg_name, arg_info in tool.inputs.items() %}{{ arg_name }}: {{ arg_info.type }}{% if not loop.last %}, {% endif %}{% endfor %}) -> {{tool.output_type}}:
    \"\"\"{{ tool.description }}

    Args:
    {%- for arg_name, arg_info in tool.inputs.items() %}
        {{ arg_name }}: {{ arg_info.description }}
    {%- endfor %}
    \"\"\"
{% endfor %}
```

The tools you have been given will provide you with access to a dataset with the following description:

Context: {{context_description}}\n

{%- if managed_agents and managed_agents.values() | list %}
You can also give tasks to team members.
Calling a team member works similarly to calling a tool: provide the task description as the 'task' argument. Since this team member is a real human, be as detailed and verbose as necessary in your task description.
You can also include any relevant variables or context using the 'additional_args' argument.
Here is a list of the team members that you can call:
```python
{%- for agent in managed_agents.values() %}
def {{ agent.name }}(task: str, additional_args: dict[str, Any]) -> str:
    \"\"\"{{ agent.description }}

    Args:
        task: Long detailed description of the task.
        additional_args: Dictionary of extra inputs to pass to the managed agent, e.g. images, dataframes, or any other contextual data it may need.
    \"\"\"
{% endfor %}
```
{%- endif %}

---
Now begin! Here is your task:
```
search: {{task}}
```
First in part 1, write the facts survey, then in part 2, write your plan.
"""

DATA_DISCOVERY_AGENT_UPDATE_PLAN_PRE_MESSAGES_PROMPT = """You are a world expert at analyzing a situation, and plan accordingly towards solving a task.
You have been given the following task:
```
search: {{task}}
```
You are working with the following dataset:
```
context: {{context_description}}
```

Below you will find a history of attempts made to solve this task.
You will first have to produce a survey of known and unknown facts, then propose a step-by-step high-level plan to solve the task.
If the previous tries so far have met some success, your updated plan can build on these results.
If you are stalled, you can make a completely new plan starting from scratch.

Find the task and history below:
"""

DATA_DISCOVERY_AGENT_UPDATE_PLAN_POST_MESSAGES_PROMPT = """Now write your updated facts below, taking into account the above history:
## 1. Updated facts survey
### 1.1. Facts given in the task
### 1.2. Facts that we have learned
### 1.3. Facts still to look up
### 1.4. Facts still to derive
### 1.5. Fact Sources

Then write a step-by-step high-level plan to solve the task above.
## 2. Plan
### 2. 1. ...
Etc.
This plan should involve individual tasks based on the available tools, that if executed correctly will yield the correct answer.
Beware that you have {remaining_steps} steps remaining.
Do not skip steps, do not add any superfluous steps. Only write the high-level plan, DO NOT DETAIL INDIVIDUAL TOOL CALLS.
After writing the final step of the plan, write the '<end_plan>' tag and stop there.

You can leverage these tools, behaving like regular python functions:
```python
{%- for tool in tools.values() %}
def {{ tool.name }}({% for arg_name, arg_info in tool.inputs.items() %}{{ arg_name }}: {{ arg_info.type }}{% if not loop.last %}, {% endif %}{% endfor %}) -> {{tool.output_type}}:
    \"\"\"{{ tool.description }}

    Args:
    {%- for arg_name, arg_info in tool.inputs.items() %}
        {{ arg_name }}: {{ arg_info.description }}
    {%- endfor %}\"\"\"
{% endfor %}
```

The tools you have been given will provide you with access to a dataset with the following description:

Context: {{context_description}}

{%- if managed_agents and managed_agents.values() | list %}
You can also give tasks to team members.
Calling a team member works similarly to calling a tool: provide the task description as the 'task' argument. Since this team member is a real human, be as detailed and verbose as necessary in your task description.
You can also include any relevant variables or context using the 'additional_args' argument.
Here is a list of the team members that you can call:
```python
{%- for agent in managed_agents.values() %}
def {{ agent.name }}(task: str, additional_args: dict[str, Any]) -> str:
    \"\"\"{{ agent.description }}

    Args:
        task: Long detailed description of the task.
        additional_args: Dictionary of extra inputs to pass to the managed agent, e.g. images, dataframes, or any other contextual data it may need.
    \"\"\"
{% endfor %}
```
{%- endif %}

Now write your updated facts survey below, then your new plan.
"""

DATA_DISCOVERY_AGENT_TASK_PROMPT = """You're a helpful agent named '{{name}}'.
You have been submitted this task by your manager.
---
Task:
{{task}}
---
You have also been given access to tools which will help you navigate a dataset with the following description:
---
Context:
{{context_description}}
---
You're helping your manager solve a search task: so make sure to not provide a one-line answer, but give as much information as possible to give them a clear understanding of the answer and the context in which you produced it. In particular, be sure to report what source(s) you use and what the contents of those source(s) contain, even if they are irrelevant for solving this task.

Your final_answer WILL HAVE to contain these parts:
### 1. Task outcome (short version):
### 2. Task outcome (extremely detailed version):
### 3. Additional context (if relevant):

Put all these in your final_answer tool, everything that you do not pass as an argument to final_answer will be lost.
And even if your task resolution is not successful, please return as much context as possible, so that your manager can act upon this feedback.
"""

DATA_DISCOVERY_AGENT_REPORT_PROMPT = """Here is the final answer from your managed agent '{{name}}':
{{final_answer}}"""

FINAL_ANSWER_PRE_MESSAGES_PROMPT = """An agent tried to answer a user query but it got stuck and failed to do so. You are tasked with providing an answer instead. Here is the agent's memory:"""

FINAL_ANSWER_POST_MESSAGES_PROMPT = """Based on the above, please provide an answer to the following user task:
{{task}}
"""



from palimpzest.prompts.agent_prompts import (
    CODE_AGENT_SYSTEM_PROMPT,
    DATA_DISCOVERY_AGENT_INITIAL_PLAN_PROMPT,
    DATA_DISCOVERY_AGENT_REPORT_PROMPT,
    DATA_DISCOVERY_AGENT_TASK_PROMPT,
    DATA_DISCOVERY_AGENT_UPDATE_PLAN_POST_MESSAGES_PROMPT,
    DATA_DISCOVERY_AGENT_UPDATE_PLAN_PRE_MESSAGES_PROMPT,
    FINAL_ANSWER_POST_MESSAGES_PROMPT,
    FINAL_ANSWER_PRE_MESSAGES_PROMPT,
)
from palimpzest.prompts.context_search import CONTEXT_SEARCH_PROMPT
from palimpzest.prompts.prompt_factory import PromptFactory
from palimpzest.prompts.utils import (
    ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION,
    ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION,
)
from palimpzest.prompts.validator import (
    FLAT_MAP_IMAGE_VALIDATOR_PROMPT,
    FLAT_MAP_VALIDATOR_PROMPT,
    MAP_IMAGE_VALIDATOR_PROMPT,
    MAP_VALIDATOR_PROMPT,
    RETRIEVE_VALIDATOR_PROMPT,
)

__all__ = [
    # agent prompts
    "CODE_AGENT_SYSTEM_PROMPT",
    "DATA_DISCOVERY_AGENT_INITIAL_PLAN_PROMPT",
    "DATA_DISCOVERY_AGENT_REPORT_PROMPT",
    "DATA_DISCOVERY_AGENT_TASK_PROMPT",
    "DATA_DISCOVERY_AGENT_UPDATE_PLAN_POST_MESSAGES_PROMPT",
    "DATA_DISCOVERY_AGENT_UPDATE_PLAN_PRE_MESSAGES_PROMPT",
    "FINAL_ANSWER_POST_MESSAGES_PROMPT",
    "FINAL_ANSWER_PRE_MESSAGES_PROMPT",
    # context search
    "CONTEXT_SEARCH_PROMPT",
    # prompt factory
    "PromptFactory",
    # utils
    "ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION",
    "ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION",
    # validator
    "FLAT_MAP_IMAGE_VALIDATOR_PROMPT",
    "FLAT_MAP_VALIDATOR_PROMPT",
    "MAP_IMAGE_VALIDATOR_PROMPT",
    "MAP_VALIDATOR_PROMPT",
    "RETRIEVE_VALIDATOR_PROMPT",
]



"""
This file contains the Generator classes and generator factory.
"""

from __future__ import annotations

import json
import logging
import os
import time
import warnings
from copy import deepcopy
from typing import Any, Generic, TypeVar

import litellm
import regex as re  # Use regex instead of re to used variable length lookbehind
from colorama import Fore, Style
from pydantic.fields import FieldInfo

from palimpzest.constants import (
    MODEL_CARDS,
    Cardinality,
    Model,
    PromptStrategy,
)
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.models import GenerationStats
from palimpzest.prompts import PromptFactory

# DEFINITIONS
GenerationOutput = tuple[dict, str | None, GenerationStats, list[dict]]
ContextType = TypeVar("ContextType")
InputType = TypeVar("InputType")


logger = logging.getLogger(__name__)

def get_json_from_answer(answer: str, model: Model, cardinality: Cardinality) -> dict[str, Any]:
    """
    This function parses an LLM response which is supposed to output a JSON object
    and optimistically searches for the substring containing the JSON object.
    """
    # model-specific trimming for LLAMA3 responses
    if model.is_llama_model():
        answer = answer.split("---")[0]
        answer = answer.replace("True", "true")
        answer = answer.replace("False", "false")

    # split off context / excess, which models sometimes output after answer
    answer = answer.split("Context:")[0]
    answer = answer.split("# this is the answer")[0]

    # trim the answer to only include the JSON dictionary
    if cardinality == Cardinality.ONE_TO_ONE:
        if not answer.strip().startswith("{"):
            # Find the start index of the actual JSON string assuming the prefix is followed by the JSON dictionary
            start_index = answer.find("{")
            if start_index != -1:
                # Remove the prefix and any leading characters before the JSON starts
                answer = answer[start_index:]

        if not answer.strip().endswith("}"):
            # Find the end index of the actual JSON string assuming the suffix is preceded by the JSON dictionary
            end_index = answer.rfind("}")
            if end_index != -1:
                # Remove the suffix and any trailing characters after the JSON ends
                answer = answer[: end_index + 1]

    # otherwise, trim the answer to only include the JSON array
    else:
        if not answer.strip().startswith("["):
            # Find the start index of the actual JSON string assuming the prefix is followed by the JSON array
            start_index = answer.find("[")
            if start_index != -1:
                # Remove the prefix and any leading characters before the JSON starts
                answer = answer[start_index:]

        if not answer.strip().endswith("]"):
            # Find the end index of the actual JSON string
            # assuming the suffix is preceded by the JSON object/array
            end_index = answer.rfind("]")
            if end_index != -1:
                # Remove the suffix and any trailing characters after the JSON ends
                answer = answer[: end_index + 1]

    # Handle weird escaped values. I am not sure why the model
    # is returning these, but the JSON parser can't take them
    answer = answer.replace(r"\_", "_")
    answer = answer.replace("\\n", "\n")
    # Remove https and http prefixes to not conflict with comment detection
    # Handle comments in the JSON response. Use regex from // until end of line
    answer = re.sub(r"(?<!https?:)\/\/.*?$", "", answer, flags=re.MULTILINE)
    answer = re.sub(r",\n.*\.\.\.$", "", answer, flags=re.MULTILINE)
    # Sanitize newlines in the JSON response
    answer = answer.replace("\n", " ")

    # finally, parse and return the JSON object; errors are handled by the caller
    return json.loads(answer)

# TODO: push parallelism of generations into LiteLLM rather than threadpool in executor
# TODO: make sure answer parsing works with custom prompts / parsers (can defer this)
class Generator(Generic[ContextType, InputType]):
    """
    Class for generating new fields for a record using an LLM.
    """

    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy,
        reasoning_effort: str | None,
        api_base: str | None = None,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        desc: str | None = None,
        verbose: bool = False,
    ):
        self.model = model
        self.model_name = model.value
        self.cardinality = cardinality
        self.prompt_strategy = prompt_strategy
        self.reasoning_effort = reasoning_effort
        self.api_base = api_base
        self.desc = desc
        self.verbose = verbose
        self.prompt_factory = PromptFactory(prompt_strategy, model, cardinality, desc)

    def _parse_reasoning(self, completion_text: str, **kwargs) -> str:
        """Extract the reasoning for the generated output from the completion object."""
        # use a custom reasoning parser if provided
        if kwargs.get("parse_reasoning"):
            parse_reasoning_fn = kwargs.get("parse_reasoning")
            return parse_reasoning_fn(completion_text)

        # if the model followed the default instructions, the completion text will have reasoning
        # before the "ANSWER:"; if this is the case, we simply extract and return that full section
        if "answer" in completion_text.lower():
            regex = re.compile("(.*?)answer:.*", re.IGNORECASE | re.DOTALL)
            matches = regex.findall(completion_text)
            if len(matches) > 0:
                return matches[0].strip()

        # otherwise, return the full completion text
        return completion_text

    def _prepare_field_answers(self, field_answers: dict | list[dict], fields: dict[str, FieldInfo]) -> dict[str, list]:
        """
        field_answers is a dictionary mapping fields to their values. For one-to-one converts, wrap each
        answer in a list. For one-to-many converts, invert the list of dictionaries into a dictionary with
        list values.
        """
        # if this is a one-to-one convert, we need to wrap each answer in a list
        if self.cardinality == Cardinality.ONE_TO_ONE:
            field_answers = {field_name: [field_answers[field_name]] for field_name in fields}

        # otherwise, we need to invert the list of dictionaries into a dictionary with list values
        else:
            field_answers_lst: list[dict] = deepcopy(field_answers)

            field_answers = {field_name: [] for field_name in fields}
            for answer_dict in field_answers_lst:
                for field_name in fields:
                    answer = answer_dict.get(field_name, None)
                    field_answers[field_name].append(answer)

        return field_answers

    def _check_convert_answer_text(self, answer_text: str, fields: dict[str, FieldInfo], throw_exception: bool=False) -> dict | list[dict] | None:
        """
        Try parsing the answer text into a JSON object. If the parsing fails, return None.
        """
        try:
            # extract json from the answer text
            field_answers = get_json_from_answer(answer_text, self.model, self.cardinality)

            # prepare the field answers to match the expected output and return
            return self._prepare_field_answers(field_answers, fields)

        except Exception as e:
            if throw_exception:
                raise e

        return None

    def _check_bool_answer_text(self, answer_text: str, throw_exception: bool=False) -> dict | None:
        """
        Return {"passed_operator": True} if and only if "true" is in the answer text.
        Return {"passed_operator": False} if and only if "false" is in the answer text.
        Otherwise, raise an exception.
        """
        # NOTE: we may be able to eliminate this condition by specifying this JSON output in the prompt;
        # however, that would also need to coincide with a change to allow the parse_answer_fn to set "passed_operator"
        if "true" in answer_text.lower():
            return {"passed_operator": True}
        elif "false" in answer_text.lower():
            return {"passed_operator": False}

        if throw_exception:
            raise Exception(f"Could not parse answer from completion text: {answer_text}")

        return None

    def _parse_convert_answer(self, completion_text: str, fields: dict[str, FieldInfo], json_output: bool) -> dict[str, list]:
        """Extract the answer from the completion object for convert operations."""
        # if the model followed the default instructions, the completion text will place
        # its answer between "ANSWER:" and "---"
        regex = re.compile("answer:(.*?)---", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()

            # if we don't expect a JSON output, return the answer text as is
            if not json_output:
                return answer_text

            # otherwise, try to parse the answer text into a JSON object
            field_answers = self._check_convert_answer_text(answer_text, fields)
            if field_answers is not None:
                return field_answers

        # if the first regex didn't find an answer, try taking all the text after "ANSWER:"
        regex = re.compile("answer:(.*)", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()

            # if we don't expect a JSON output, return the answer text as is
            if not json_output:
                return answer_text
            
            # otherwise, try to parse the answer text into a JSON object
            field_answers = self._check_convert_answer_text(answer_text, fields)
            if field_answers is not None:
                return field_answers

        # finally, try taking all of the text; for JSON output, throw an exception if parsing fails
        if not json_output:
            return completion_text

        return self._check_convert_answer_text(completion_text, fields, throw_exception=True)

    def _parse_bool_answer(self, completion_text: str, json_output: bool) -> dict[str, list]:
        """Extract the answer from the completion object for filter and join operations."""
        # if the model followed the default instructions, the completion text will place
        # its answer between "ANSWER:" and "---"
        regex = re.compile("answer:(.*?)---", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()

            # if we don't expect a JSON output, return the answer text as is
            if not json_output:
                return answer_text

            # otherwise, try to parse the answer text into a JSON object
            field_answers = self._check_bool_answer_text(answer_text)
            if field_answers is not None:
                return field_answers

        # if the first regex didn't find an answer, try taking all the text after "ANSWER:"
        regex = re.compile("answer:(.*)", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()

            # if we don't expect a JSON output, return the answer text as is
            if not json_output:
                return answer_text

            # otherwise, try to parse the answer text into a JSON object
            field_answers = self._check_bool_answer_text(answer_text)
            if field_answers is not None:
                return field_answers

        # finally, try taking all of the text; for JSON output, throw an exception if parsing fails
        if not json_output:
            return completion_text

        return self._check_bool_answer_text(completion_text, throw_exception=True)

    def _parse_answer(self, completion_text: str, fields: dict[str, FieldInfo] | None, json_output: bool, **kwargs) -> dict[str, list]:
        """Extract the answer from the completion object."""
        # use a custom answer parser if provided
        if kwargs.get("parse_answer"):
            parse_answer_fn = kwargs.get("parse_answer")
            return parse_answer_fn(completion_text)

        # fields should be a dict if a custom answer parser is not provided
        assert isinstance(fields, dict), "Fields must be provided if a custom answer parser is not provided."

        # extract the per-field answers from the completion text
        field_answers = (
            self._parse_bool_answer(completion_text, json_output)
            if self.prompt_strategy.is_filter_prompt() or self.prompt_strategy.is_join_prompt()
            else self._parse_convert_answer(completion_text, fields, json_output)
        )

        return field_answers

    def __call__(self, candidate: DataRecord | list[DataRecord], fields: dict[str, FieldInfo] | None, right_candidate: DataRecord | None = None, json_output: bool=True, **kwargs) -> GenerationOutput:
        """Take the input record(s) (`candidate`), generate the output `fields`, and return the generated output."""
        logger.debug(f"Generating for candidate(s) {candidate} with fields {fields}")

        # fields can only be None if the user provides an answer parser
        fields_check = fields is not None or "parse_answer" in kwargs
        assert fields_check, "`fields` must be provided if `parse_answer` function is not provided in kwargs."

        # if the user (or operator) provides a system prompt instead of a prompt, treat this as
        # the prompt and print a warning
        if "system_prompt" in kwargs and "prompt" not in kwargs:
            kwargs["prompt"] = kwargs["system_prompt"]
            kwargs.pop("system_prompt")
            warnings.warn("Provided `system_prompt` without providing `prompt`; setting `prompt` = `system_prompt`.")  # noqa: B028

        # generate a list of messages which can be used to construct a payload
        messages = self.prompt_factory.create_messages(candidate, fields, right_candidate, **kwargs)
        is_audio_op = any(msg.get("type") == "input_audio" for msg in messages)

        # generate the text completion
        start_time = time.time()
        completion = None
        try:
            completion_kwargs = {}
            if not self.model.is_o_model() and not self.model.is_gpt_5_model():
                completion_kwargs = {"temperature": kwargs.get("temperature", 0.0), **completion_kwargs}
            if is_audio_op:
                completion_kwargs = {"modalities": ["text"], **completion_kwargs}
            if self.model.is_reasoning_model():
                completion_kwargs = {"reasoning_effort": self.reasoning_effort, **completion_kwargs}
            if self.model.is_vllm_model():
                completion_kwargs = {"api_base": self.api_base, "api_key": os.environ.get("VLLM_API_KEY", "fake-api-key"), **completion_kwargs}
            completion = litellm.completion(model=self.model_name, messages=messages, **completion_kwargs)
            end_time = time.time()
            logger.debug(f"Generated completion in {end_time - start_time:.2f} seconds")
        # if there's an error generating the completion, we have to return an empty answer
        # and can only account for the time spent performing the failed generation
        except Exception as e:
            logger.error(f"Error generating completion: {e}")
            field_answers = (
                {"passed_operator": False}
                if self.prompt_strategy.is_filter_prompt() or self.prompt_strategy.is_join_prompt()
                else {field_name: None for field_name in fields}
            )
            reasoning = None
            generation_stats = GenerationStats(
                model_name=self.model_name,
                llm_call_duration_secs=time.time() - start_time,
                total_llm_calls=1,
            )

            return field_answers, reasoning, generation_stats, messages

        # parse usage statistics and create the GenerationStats
        generation_stats = None
        if completion is not None:
            usage = completion.usage.model_dump()

            # get cost per input/output token for the model
            usd_per_input_token = MODEL_CARDS[self.model_name].get("usd_per_input_token", 0.0)
            usd_per_audio_input_token = MODEL_CARDS[self.model_name].get("usd_per_audio_input_token", 0.0)
            usd_per_output_token = MODEL_CARDS[self.model_name]["usd_per_output_token"]

            # TODO: for some models (e.g. GPT-5) we cannot separate text from image prompt tokens yet;
            #       for now, we only use tokens from prompt_token_details if it's an audio prompt
            # get output tokens (all text) and input tokens by modality
            output_tokens = usage["completion_tokens"]
            if is_audio_op:
                input_audio_tokens = usage["prompt_tokens_details"].get("audio_tokens", 0)
                input_text_tokens = usage["prompt_tokens_details"].get("text_tokens", 0)
                input_image_tokens = 0
            else:
                input_audio_tokens = 0
                input_text_tokens = usage["prompt_tokens"]
                input_image_tokens = 0
            input_tokens = input_audio_tokens + input_text_tokens + input_image_tokens

            # compute the input and output token costs
            total_input_cost = (input_text_tokens + input_image_tokens) * usd_per_input_token + input_audio_tokens * usd_per_audio_input_token
            total_output_cost = output_tokens * usd_per_output_token

            generation_stats = GenerationStats(
                model_name=self.model_name,
                llm_call_duration_secs=end_time - start_time,
                fn_call_duration_secs=0.0,
                input_audio_tokens=input_audio_tokens,
                input_text_tokens=input_text_tokens,
                input_image_tokens=input_image_tokens,
                total_input_tokens=input_tokens,
                total_output_tokens=output_tokens,
                total_input_cost=total_input_cost,
                total_output_cost=total_output_cost,
                cost_per_record=total_input_cost + total_output_cost,
                total_llm_calls=1,
            )

        # pretty print prompt + full completion output for debugging
        completion_text = completion.choices[0].message.content
        prompt, system_prompt = "", ""
        for message in messages:
            if message["role"] == "system":
                system_prompt += message["content"] + "\n"
            if message["role"] == "user":
                if message["type"] == "text":
                    prompt += message["content"] + "\n"
                elif message["type"] == "image":
                    prompt += "<image>\n" * len(message["content"])
                elif message["type"] == "input_audio":
                    prompt += "<audio>\n" * len(message["content"])
        logger.debug(f"PROMPT:\n{prompt}")
        logger.debug(Fore.GREEN + f"{completion_text}\n" + Style.RESET_ALL)

        # parse reasoning
        reasoning = None
        try:
            reasoning = self._parse_reasoning(completion_text, **kwargs)
        except Exception as e:
            logger.error(f"Error parsing reasoning and answers: {e}")
            pass

        # parse field answers
        field_answers = None 
        if fields is not None and (self.prompt_strategy.is_filter_prompt() or self.prompt_strategy.is_join_prompt()):
            field_answers = {"passed_operator": False}
        elif fields is not None and not (self.prompt_strategy.is_filter_prompt() or self.prompt_strategy.is_join_prompt()):
            field_answers = {field_name: None for field_name in fields}
        try:
            field_answers = self._parse_answer(completion_text, fields, json_output, **kwargs)
        except Exception as e:
            logger.error(f"Error parsing answers: {e}")
            os.makedirs("parse-answer-errors", exist_ok=True)
            ts = time.time()
            with open(f"parse-answer-errors/error-{ts}.txt", "w") as f:
                f.write(f"{str(self.model_name)}\n")
                f.write("#####\n")
                f.write(f"{str(self.prompt_strategy)}\n")
                f.write("#####\n")
                f.write(f"{str(completion_text)}\n")
                f.write("#####\n")
                f.write(f"{str(fields)}\n")
                f.write("#####\n")
                f.write(f"{str(e)}\n")

        logger.debug(f"Generated field answers: {field_answers}")
        return field_answers, reasoning, generation_stats, messages



import logging

import numpy as np

from palimpzest.core.data.dataset import Dataset
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import SentinelPlanStats
from palimpzest.query.execution.execution_strategy import SentinelExecutionStrategy
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.filter import FilterOp
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ContextScanOp, ScanPhysicalOp
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.utils.progress import create_progress_manager
from palimpzest.validator.validator import Validator

logger = logging.getLogger(__name__)

class OpSet:
    """
    This class represents the set of operators which are currently in the frontier for a given logical operator.
    Each operator in the frontier is an instance of a PhysicalOperator which either:

    1. lies on the Pareto frontier of the set of sampled operators, or
    2. has been sampled fewer than j times
    """

    def __init__(self, op_set: list[PhysicalOperator], source_unique_logical_op_ids: list[str], source_indices: list[int]):
        # construct the set of operators
        self.ops = op_set

        # store the order in which we will sample the source records
        self.source_indices = source_indices

        # boolean indication of the type of operator in this OpSet
        sample_op = op_set[0]
        self.is_scan_op = isinstance(sample_op, (ScanPhysicalOp, ContextScanOp))
        self.is_filter_op = isinstance(sample_op, FilterOp)
        self.is_aggregate_op = isinstance(sample_op, AggregateOp)
        self.is_llm_join = isinstance(sample_op, JoinOp)

        # set the initial inputs for this logical operator
        self.source_indices_to_inputs = {source_unique_logical_op_id: {} for source_unique_logical_op_id in source_unique_logical_op_ids}
        if self.is_scan_op:
            self.source_indices_to_inputs["source"] = {source_idx: [int(source_idx.split("-")[-1])] for source_idx in self.source_indices}

    def get_op_inputs(self) -> list[PhysicalOperator, DataRecord | int | None]:
        """
        Returns the list of frontier operators and their next input to process.
        """
        # if this is an aggregate, run on every input
        if self.is_aggregate_op:
            op = self.ops[0]
            all_inputs = []
            for _, source_indices_to_inputs in self.source_indices_to_inputs.items():
                for _, inputs in source_indices_to_inputs.items():
                    all_inputs.extend(inputs)
            return [(op, tuple(), all_inputs)]

        # if this is an un-optimized (non-scan, non-join) operator, flatten inputs and run on each one
        elif not self.is_scan_op and not self.is_llm_join and len(self.ops) == 1:
            op_inputs = []
            op = self.ops[0]
            for _, source_indices_to_inputs in self.source_indices_to_inputs.items():
                for source_indices, inputs in source_indices_to_inputs.items():
                    for input in inputs:
                        op_inputs.append((op, source_indices, input))
            return op_inputs

        # get the list of (op, source_indices) pairs which this operator needs to execute
        op_source_indices_pairs = []
        for op in self.ops:
            # construct list of inputs by looking up the input for the given source_indices
            for source_indices in self.source_indices:
                op_source_indices_pairs.append((op, source_indices))

        # construct the op inputs
        op_inputs = []
        if self.is_llm_join:
            left_source_unique_logical_op_id, right_source_unique_logical_op_id = list(self.source_indices_to_inputs)
            left_source_indices_to_inputs = self.source_indices_to_inputs[left_source_unique_logical_op_id]
            right_source_indices_to_inputs = self.source_indices_to_inputs[right_source_unique_logical_op_id]
            for op, source_indices in op_source_indices_pairs:
                left_source_indices = source_indices[0]
                right_source_indices = source_indices[1]
                left_inputs = left_source_indices_to_inputs.get(left_source_indices, [])
                right_inputs = right_source_indices_to_inputs.get(right_source_indices, [])
                if len(left_inputs) > 0 and len(right_inputs) > 0:
                    op_inputs.append((op, (left_source_indices, right_source_indices), (left_inputs, right_inputs)))
            return op_inputs

        # if operator is not a join
        source_unique_logical_op_id = list(self.source_indices_to_inputs)[0]
        op_inputs = [
            (op, source_indices, input)
            for op, source_indices in op_source_indices_pairs
            for input in self.source_indices_to_inputs[source_unique_logical_op_id].get(source_indices, [])
        ]

        return op_inputs

    def pick_highest_quality_output(self, record_sets: list[DataRecordSet]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(record_sets) == 1:
            return record_sets[0]

        # NOTE: I don't like that this assumes the models are consistent in
        #       how they order their record outputs for one-to-many converts;
        #       eventually we can try out more robust schemes to account for
        #       differences in ordering
        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set in record_sets:
            for idx in range(len(record_set)):
                record, record_op_stats = record_set[idx], record_set.record_op_stats[idx]
                if idx not in idx_to_records:
                    idx_to_records[idx] = [(record, record_op_stats)]
                else:
                    idx_to_records[idx].append((record, record_op_stats))

        # compute highest quality answer at each index
        out_records = []
        out_record_op_stats = []
        for idx in range(len(idx_to_records)):
            records_lst, record_op_stats_lst = zip(*idx_to_records[idx])
            max_quality_record, max_quality = records_lst[0], record_op_stats_lst[0].quality
            max_quality_stats = record_op_stats_lst[0]
            for record, record_op_stats in zip(records_lst[1:], record_op_stats_lst[1:]):
                record_quality = record_op_stats.quality
                if record_quality > max_quality:
                    max_quality_record = record
                    max_quality = record_quality
                    max_quality_stats = record_op_stats
            out_records.append(max_quality_record)
            out_record_op_stats.append(max_quality_stats)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, out_record_op_stats)

    def update_inputs(self, source_idx_to_record_sets: dict[int, DataRecordSet]):
        """
        Update the inputs for this logical operator based on the outputs of the previous logical operator.
        """
        for source_idx, record_sets in source_idx_to_record_sets.items():
            input = []
            max_quality_record_set = self.pick_highest_quality_output(record_sets)
            for record in max_quality_record_set:
                input.append(record if record._passed_operator else None)

            self.source_indices_to_inputs[source_idx] = input

class AllSamplingExecutionStrategy(SentinelExecutionStrategy):

    def _execute_sentinel_plan(self,
            plan: SentinelPlan,
            op_sets: dict[str, OpSet],
            validator: Validator,
            plan_stats: SentinelPlanStats,
        ) -> SentinelPlanStats:
        # execute operator sets in sequence
        for topo_idx, (logical_op_id, _) in enumerate(plan):
            # compute unique logical op id within plan
            unique_logical_op_id = f"{topo_idx}-{logical_op_id}"

            # get frontier ops and their next input
            op_inputs = op_sets[logical_op_id].get_op_inputs()

            # break out of the loop if op_inputs is empty, as this means all records have been filtered out
            if len(op_inputs) == 0:
                break

            # run sampled operators on sampled inputs
            source_indices_to_record_set_tuples, _ = self._execute_op_set(unique_logical_op_id, op_inputs)

            # score the quality of each generated output
            source_indices_to_all_record_sets = {
                    source_indices: [(record_set, op) for record_set, op, _ in record_set_tuples]
                    for source_indices, record_set_tuples in source_indices_to_record_set_tuples.items()
                }
            source_indices_to_all_record_sets, val_gen_stats = self._score_quality(validator, source_indices_to_all_record_sets)

            # remove records that were read from the execution cache before adding to record op stats
            new_record_op_stats = []
            for _, record_set_tuples in source_indices_to_record_set_tuples.items():
                for record_set, _, is_new in record_set_tuples:
                    if is_new:
                        new_record_op_stats.extend(record_set.record_op_stats)

            # update plan stats
            plan_stats.add_record_op_stats(unique_logical_op_id, new_record_op_stats)
            plan_stats.add_validation_gen_stats(unique_logical_op_id, val_gen_stats)

            # provide the best record sets as inputs to the next logical operator
            next_unique_logical_op_id = plan.get_next_unique_logical_op_id(unique_logical_op_id)
            if next_unique_logical_op_id is not None:
                source_indices_to_all_record_sets = {
                    source_indices: [record_set for record_set, _ in record_set_tuples]
                    for source_indices, record_set_tuples in source_indices_to_all_record_sets.items()
                }
                op_sets[next_unique_logical_op_id].update_inputs(unique_logical_op_id, source_indices_to_all_record_sets)

        # finalize plan stats
        plan_stats.finish()

        return plan_stats

    def execute_sentinel_plan(self, plan: SentinelPlan, train_dataset: dict[str, Dataset], validator: Validator): # expected_outputs: dict[int, dict] | None):
        """
        NOTE: this function currently requires us to set k and j properly in order to make
              comparison in our research against the corresponding sample budget in MAB.

        NOTE: the number of samples will slightly exceed the sample_budget if the number of operator
        calls does not perfectly match the sample_budget. This may cause some minor discrepancies with
        the progress manager as a result.
        """
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = SentinelPlanStats.from_plan(plan)
        plan_stats.start()

        # get lists of source indices
        dataset_id_to_source_indices = {}
        for dataset_id, dataset in train_dataset.items():
            total_num_samples = len(dataset)
            source_indices = [f"{dataset_id}---{int(idx)}" for idx in np.arange(total_num_samples)]
            dataset_id_to_source_indices[dataset_id] = source_indices

        # initialize set of physical operators for each logical operator
        op_sets = {}
        for topo_idx, (logical_op_id, op_set) in enumerate(plan):
            unique_logical_op_id = f"{topo_idx}-{logical_op_id}"
            source_unique_logical_op_ids = plan.get_source_unique_logical_op_ids(unique_logical_op_id)
            sample_op = op_set[0]
            if isinstance(sample_op, (ScanPhysicalOp, ContextScanOp)):
                root_dataset_ids = plan.get_root_dataset_ids(unique_logical_op_id)
                assert len(root_dataset_ids) == 1, f"Scan for {sample_op} has {len(root_dataset_ids)} > 1 root dataset ids"
                root_dataset_id = root_dataset_ids[0]
                source_indices = dataset_id_to_source_indices[root_dataset_id]
                op_sets[unique_logical_op_id] = OpSet(op_set, source_unique_logical_op_ids, source_indices)
            elif isinstance(sample_op, JoinOp):
                assert len(source_unique_logical_op_ids) == 2, f"Join for {sample_op} has {len(source_unique_logical_op_ids)} != 2 source logical operators"
                left_source_indices = op_sets[source_unique_logical_op_ids[0]].source_indices
                right_source_indices = op_sets[source_unique_logical_op_ids[1]].source_indices
                source_indices = []
                for left_source_idx in left_source_indices:
                    for right_source_idx in right_source_indices:
                        source_indices.append((left_source_idx, right_source_idx))
                op_sets[unique_logical_op_id] = OpSet(op_set, source_unique_logical_op_ids, source_indices)
            else:
                source_indices = op_sets[source_unique_logical_op_ids[0]].source_indices
                op_sets[unique_logical_op_id] = OpSet(op_set, source_unique_logical_op_ids, source_indices)

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, sample_budget=self.sample_budget, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _execute_sentinel_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail because
        #       the progress manager cannot get a handle to the console 
        try:
            # execute sentinel plan by sampling records and operators
            plan_stats = self._execute_sentinel_plan(plan, op_sets, validator, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing sentinel plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return plan_stats



from __future__ import annotations

from abc import ABC, abstractmethod

from palimpzest.core.models import PlanCost
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ContextScanOp, MarshalAndScanDataOp
from palimpzest.utils.hash_helpers import hash_for_id


class Plan(ABC):
    @abstractmethod
    def compute_plan_id(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other) -> bool:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __getitem__(self, slice) -> tuple:
        pass

    @abstractmethod
    def __iter__(self) -> iter:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

class PhysicalPlan(Plan):
    def __init__(self, operator: PhysicalOperator, subplans: list[PhysicalPlan] | None, plan_cost: PlanCost | None = None):
        self.operator = operator
        self.subplans = [] if subplans is None else subplans
        self.plan_cost = plan_cost if plan_cost is not None else PlanCost(cost=0.0, time=0.0, quality=1.0)
        self.plan_id = self.compute_plan_id()

        # NOTE: unique full_op_id is constructed as "{topological_index}-{full_op_id}" to
        # differentiate between multiple instances of the same physical operator e.g. in self-joins

        # compute mapping from unique full_op_id to next unique full_op_id in the plan
        self.unique_full_op_id_to_next_unique_full_op_and_id = {}
        current_idx, _ = self._compute_next_unique_full_op_map(self.unique_full_op_id_to_next_unique_full_op_and_id)
        self.unique_full_op_id_to_next_unique_full_op_and_id[f"{current_idx}-{self.operator.get_full_op_id()}"] = (None, None)

        # compute mapping from unique full_op_id to upstream unique full_op_ids
        self.unique_full_op_id_to_upstream_full_op_ids = {}
        self._compute_upstream_unique_full_op_ids_map(self.unique_full_op_id_to_upstream_full_op_ids)

        # compute mapping from unique full_op_id to source unique full_op_ids
        self.unique_full_op_id_to_source_full_op_ids = {}
        self._compute_source_unique_full_op_ids_map(self.unique_full_op_id_to_source_full_op_ids)

    def compute_plan_id(self) -> str:
        """
        NOTE: This is NOT a universal ID.

        Two different PhysicalPlan instances with the identical lists of operators will have equivalent plan_ids.
        """
        full_op_id = self.operator.get_full_op_id()
        subplan_ids = [subplan.compute_plan_id() for subplan in self.subplans]
        return hash_for_id(str((full_op_id,) + tuple(subplan_ids)))

    def get_est_total_outputs(self, num_samples: int | None = None, current_idx: int | None = None, source_unique_full_op_ids_map: dict | None = None) -> tuple[dict[str, int], int]:
        """Return the estimated total number of output records to be processed by the given operator in this plan."""
        # get the source map from the root of the entire plan; use this map throughout all recursive calls
        # (if you call self.get_source_unique_full_op_ids() from a subplan, it's topo indexes will be different)
        if source_unique_full_op_ids_map is None:
            source_unique_full_op_ids_map = self.unique_full_op_id_to_source_full_op_ids

        # get the estimated total outputs from all subplans
        # NOTE: this will be an empty dictionary for scans
        all_subplan_total_outputs = {}
        for subplan in self.subplans:
            subplan_total_outputs, current_idx = subplan.get_est_total_outputs(num_samples, current_idx, source_unique_full_op_ids_map)
            current_idx += 1
            all_subplan_total_outputs.update(subplan_total_outputs)

        # if current_idx is None, this is the first call, so we initialize it to 0
        if current_idx is None:
            current_idx = 0

        # get total outputs for this operator
        this_op_total_outputs = {}
        this_unique_full_op_id = f"{current_idx}-{self.operator.get_full_op_id()}"

        # if this operator is a scan, return the length of its datasource
        if isinstance(self.operator, MarshalAndScanDataOp):
            total = min(len(self.operator.datasource), num_samples) if num_samples is not None else len(self.operator.datasource)
            this_op_total_outputs = {this_unique_full_op_id: total}

        # if this operator is a context scan, return 1
        elif isinstance(self.operator, ContextScanOp):  # noqa: SIM114
            this_op_total_outputs = {this_unique_full_op_id: 1}

        # if this operator is an aggregate, return 1
        elif isinstance(self.operator, AggregateOp):
            this_op_total_outputs = {this_unique_full_op_id: 1}

        # if this operator is a limit scan, return its limit
        elif isinstance(self.operator, LimitScanOp):
            this_op_total_outputs = {this_unique_full_op_id: self.operator.limit}

        # if this operator is a join, return the Cartesian product of the estimated outputs of its inputs
        elif isinstance(self.operator, JoinOp):
            # get estimated outputs for immediate left and right inputs
            source_unique_full_op_ids = source_unique_full_op_ids_map[f"{current_idx}-{self.operator.get_full_op_id()}"]
            left_unique_full_op_id, right_unique_full_op_id = source_unique_full_op_ids[0], source_unique_full_op_ids[1]
            left_total_outputs = all_subplan_total_outputs[left_unique_full_op_id]
            right_total_outputs = all_subplan_total_outputs[right_unique_full_op_id]
            this_op_total_outputs = {this_unique_full_op_id: left_total_outputs * right_total_outputs}

        # otherwise, return the number of outputs from the immediate input
        else:
            source_unique_full_op_ids = source_unique_full_op_ids_map[f"{current_idx}-{self.operator.get_full_op_id()}"]
            source_unique_full_op_id = source_unique_full_op_ids[0]
            this_op_total_outputs = {this_unique_full_op_id: all_subplan_total_outputs[source_unique_full_op_id]}

        return {**this_op_total_outputs, **all_subplan_total_outputs}, current_idx

    def _compute_next_unique_full_op_map(self, next_map: dict[str, str | None], current_idx: int | None = None) -> tuple[int, str]:
        """Compute a mapping from each operator's unique full_op_id to the next operator in the plan and its unique full_op_id.

        The unique full_op_id is constructed as "{topological_index}-{full_op_id}" to differentiate between
        multiple instances of the same physical operator in the plan (e.g., in self-joins).

        Args:
            next_map: A dictionary to populate with the mapping from unique full_op_id to next (operator, unique_full_op_id) pair.
            current_idx: The current topological index in the plan. If None, starts at 0.

        Returns:
            A tuple containing:
                - The current topological index after processing this plan.
                - The unique full_op_id of this plan's root operator.
        """
        # If there are subplans, compute their next maps first
        subplan_topo_idx_op_id_pairs = []
        for subplan in self.subplans:
            current_idx, current_full_op_id = subplan._compute_next_unique_full_op_map(next_map, current_idx)
            subplan_topo_idx_op_id_pairs.append((current_idx, current_full_op_id))
            current_idx += 1  # increment after processing each subplan

        # for each subplan's root operator, set its next to this plan's root operator
        for topo_idx, full_op_id in subplan_topo_idx_op_id_pairs:
            unique_op_id = f"{topo_idx}-{full_op_id}"
            this_unique_op_id = f"{current_idx}-{self.operator.get_full_op_id()}"
            next_map[unique_op_id] = (self.operator, this_unique_op_id)

        # if this is the first call, initialize current_idx
        if current_idx is None:
            current_idx = 0

        return current_idx, self.operator.get_full_op_id()

    def get_next_unique_full_op_and_id(self, topo_idx: int, operator: PhysicalOperator) -> tuple[PhysicalOperator | None, str | None]:
        """Return the next operator in the plan after the given operator, or None if it is the last operator."""
        unique_full_op_id = f"{topo_idx}-{operator.get_full_op_id()}"
        return self.unique_full_op_id_to_next_unique_full_op_and_id[unique_full_op_id]

    def get_next_unique_full_op_id(self, topo_idx: int, operator: PhysicalOperator) -> str | None:
        """Return the full_op_id of the next operator in the plan after the given operator, or None if it is the last operator."""
        unique_full_op_id = f"{topo_idx}-{operator.get_full_op_id()}"
        _, next_unique_full_op_id = self.unique_full_op_id_to_next_unique_full_op_and_id[unique_full_op_id]
        return next_unique_full_op_id

    def _compute_upstream_unique_full_op_ids_map(self, upstream_map: dict[str, list[str]], current_idx: int | None = None) -> tuple[int, str, list[str]]:
        # set the upstream unique full_op_ids for this operator
        subplan_topo_idx_upstream_unique_full_op_id_tuples = []
        for subplan in self.subplans:
            current_idx, full_op_id, subplan_upstream_unique_full_op_ids = subplan._compute_upstream_unique_full_op_ids_map(upstream_map, current_idx)
            subplan_topo_idx_upstream_unique_full_op_id_tuples.append((current_idx, full_op_id, subplan_upstream_unique_full_op_ids))
            current_idx += 1

        # if current_idx is None, this is the first call, so we initialize it to 0
        if current_idx is None:
            current_idx = 0

        # compute this operator's unique full_op_id
        this_unique_full_op_id = f"{current_idx}-{self.operator.get_full_op_id()}"

        # update the upstream_map for this operator
        upstream_map[this_unique_full_op_id] = []
        for topo_idx, full_op_id, upstream_unique_full_op_ids in subplan_topo_idx_upstream_unique_full_op_id_tuples:
            subplan_upstream_unique_full_op_ids = [f"{topo_idx}-{full_op_id}"] + upstream_unique_full_op_ids
            upstream_map[this_unique_full_op_id].extend(subplan_upstream_unique_full_op_ids)

        # return the current index and the upstream unique full_op_ids for this operator
        return current_idx, self.operator.get_full_op_id(), upstream_map[this_unique_full_op_id]

    def get_upstream_unique_full_op_ids(self, unique_full_op_id: str) -> list[str]:
        """Return the list of unique full_op_ids for the upstream operators of the operator specified by `unique_full_op_id`."""
        return self.unique_full_op_id_to_upstream_full_op_ids[unique_full_op_id]

    def _compute_source_unique_full_op_ids_map(self, source_map: dict[str, list[str]], current_idx: int | None = None) -> tuple[int, str]:
        # get the topological index and full_op_id pairs for all subplans' root operators
        subplan_topo_idx_op_id_pairs = []
        for subplan in self.subplans:
            current_idx, current_full_op_id = subplan._compute_source_unique_full_op_ids_map(source_map, current_idx)
            subplan_topo_idx_op_id_pairs.append((current_idx, current_full_op_id))
            current_idx += 1

        # if current_idx is None, this is the first call, so we initialize it to 0
        if current_idx is None:
            current_idx = 0

        # compute this operator's unique full_op_id
        this_unique_full_op_id = f"{current_idx}-{self.operator.get_full_op_id()}"

        # update the source_map for this operator
        source_map[this_unique_full_op_id] = []
        for topo_idx, full_op_id in subplan_topo_idx_op_id_pairs:
            unique_full_op_id = f"{topo_idx}-{full_op_id}"
            source_map[this_unique_full_op_id].append(unique_full_op_id)

        # return the current unique full_op_id for this operator
        return current_idx, self.operator.get_full_op_id()

    def get_source_unique_full_op_ids(self, topo_idx: int, operator: PhysicalOperator) -> list[str]:
        """Return the list of unique full_op_ids for the input(s) to this operator."""
        unique_full_op_id = f"{topo_idx}-{operator.get_full_op_id()}"
        return self.unique_full_op_id_to_source_full_op_ids[unique_full_op_id]

    def __eq__(self, other):
        return isinstance(other, PhysicalPlan) and self.plan_id == other.plan_id

    def __hash__(self):
        return int(self.plan_id, 16)

    def __repr__(self) -> str:
        return str(self)

    def _get_str(self, idx: int = 0, indent: int = 0) -> str:
        indent_str = " " * (indent * 2)
        plan_str = f"{indent_str}{idx}. {str(self.operator)}\n"
        for subplan in self.subplans:
            plan_str += subplan._get_str(idx=idx + 1, indent=indent + 1)

        return plan_str

    def __str__(self):
        return self._get_str()

    def __getitem__(self, slice):
        ops = [op for op in self]
        return ops[slice]

    def __iter__(self):
        for subplan in self.subplans:
            yield from subplan
        yield self.operator

    def __len__(self):
        return 1 + sum(len(subplan) for subplan in self.subplans)

    @classmethod
    def _from_ops(cls, ops: list[PhysicalOperator], plan_cost: PlanCost | None = None) -> PhysicalPlan:
        """
        NOTE: Do not use this in production code. This is a convenience method for constructing PhysicalPlans in tests.
        This method assumes a left-deep tree structure (i.e. pipeline), where each operator has at most one subplan.
        The PlanCost is applied to all subplans, thus it is not a true representation of the cost of the plan.
        """
        assert len(ops) > 0, "ops must contain at least one PhysicalOperator"

        # build the PhysicalPlan from the list of operators
        if len(ops) == 1:
            return cls(operator=ops[0], subplans=None, plan_cost=plan_cost)

        # recursively build subplans
        subplan = cls._from_ops(ops[:-1], plan_cost=plan_cost)
        return cls(operator=ops[-1], subplans=[subplan], plan_cost=plan_cost)


# TODO(?): take list[PhysicalOperator] as input, but then store OpFrontier
class SentinelPlan(Plan):
    def __init__(self, operator_set: list[PhysicalOperator], subplans: list[SentinelPlan] | None):
        # store operator_set and logical_op_id; sort operator_set internally by full_op_id
        self.operator_set = sorted(operator_set, key=lambda op: op.get_full_op_id())
        self.logical_op_id = self.operator_set[0].logical_op_id
        self.subplans = [] if subplans is None else subplans
        self.plan_id = self.compute_plan_id()

        # compute mapping from unique logical_op_id to next unique logical_op_id in the plan
        self.unique_logical_op_id_to_next_unique_logical_op_id = {}
        current_idx, _ = self._compute_next_unique_logical_op_id_map(self.unique_logical_op_id_to_next_unique_logical_op_id)
        self.unique_logical_op_id_to_next_unique_logical_op_id[f"{current_idx}-{self.logical_op_id}"] = None

        # compute mapping from unique logical_op_id to root dataset ids
        self.unique_logical_op_id_to_root_dataset_ids = {}
        self._compute_root_dataset_ids_map(self.unique_logical_op_id_to_root_dataset_ids)

        # compute mapping from unique logical_op_id to source unique logical_op_ids
        self.unique_logical_op_id_to_source_logical_op_ids = {}
        self._compute_source_unique_logical_op_ids_map(self.unique_logical_op_id_to_source_logical_op_ids)

    def compute_plan_id(self) -> str:
        """
        NOTE: This is NOT a universal ID.

        Two different SentinelPlan instances with the identical operator_sets will have equivalent plan_ids.
        """
        full_id = (self.logical_op_id,) + tuple([op.get_full_op_id() for op in self.operator_set])
        subplan_ids = [subplan.compute_plan_id() for subplan in self.subplans]
        return hash_for_id(str((full_id,) + tuple(subplan_ids)))

    def __eq__(self, other):
        return isinstance(other, SentinelPlan) and self.plan_id == other.plan_id

    def __hash__(self):
        return int(self.plan_id, 16)

    def __repr__(self) -> str:
        return str(self)

    def _get_str(self, idx: int = 0, indent: int = 0) -> str:
        indent_str = " " * (indent * 2)
        operator = self.operator_set[0]
        inner_idx_str = "" if len(self.operator_set) == 1 else f"1 - {len(self.operator_set)}."
        plan_str = f"{indent_str}{idx}.{inner_idx_str} {str(operator)}\n"
        for subplan in self.subplans:
            plan_str += subplan._get_str(idx=idx + 1, indent=indent + 1)

        return plan_str

    def __str__(self):
        return self._get_str()

    def __getitem__(self, slice):
        op_set_tuples = [op_set_tuple for op_set_tuple in self]
        return op_set_tuples[slice]

    def __iter__(self):
        for subplan in self.subplans:
            yield from subplan
        yield self.logical_op_id, self.operator_set

    def __len__(self):
        return 1 + sum(len(subplan) for subplan in self.subplans)
    
    def _compute_next_unique_logical_op_id_map(self, next_map: dict[str, str | None], current_idx: int | None = None) -> tuple[int, str]:
        """Compute a mapping from each operator's unique logical_op_id to the next operator's unique logical_op_id.

        The unique logical_op_id is constructed as "{topological_index}-{logical_op_id}" to differentiate between
        multiple instances of the same physical operator in the plan (e.g., in self-joins).

        Args:
            next_map: A dictionary to populate with the mapping from unique logical_op_id to next logical_op_id.
            current_idx: The current topological index in the plan. If None, starts at 0.

        Returns:
            A tuple containing:
                - The current topological index after processing this plan.
                - The unique logical_op_id of this plan's root logical operator.
        """
        # If there are subplans, compute their next maps first
        subplan_topo_idx_op_id_pairs = []
        for subplan in self.subplans:
            current_idx, current_logical_op_id = subplan._compute_next_unique_logical_op_id_map(next_map, current_idx)
            subplan_topo_idx_op_id_pairs.append((current_idx, current_logical_op_id))
            current_idx += 1  # increment after processing each subplan

        # for each subplan's root operator, set its next to this plan's root operator
        for topo_idx, logical_op_id in subplan_topo_idx_op_id_pairs:
            unique_logical_op_id = f"{topo_idx}-{logical_op_id}"
            this_unique_logical_op_id = f"{current_idx}-{self.logical_op_id}"
            next_map[unique_logical_op_id] = this_unique_logical_op_id

        # if this is the first call, initialize current_idx
        if current_idx is None:
            current_idx = 0

        return current_idx, self.logical_op_id

    def get_next_unique_logical_op_id(self, unique_logical_op_id: str) -> str | None:
        """Return the unique logical_op_id of the next operator in the plan after the given operator, or None if it is the last operator."""
        return self.unique_logical_op_id_to_next_unique_logical_op_id[unique_logical_op_id]

    def _compute_root_dataset_ids_map(self, root_dataset_ids_map: dict[str, list[str]], current_idx: int | None = None) -> tuple[int, list[str]]:
        # set the root dataset ids for this operator
        all_subplan_root_dataset_ids = []
        for subplan in self.subplans:
            current_idx, subplan_root_dataset_ids = subplan._compute_root_dataset_ids_map(root_dataset_ids_map, current_idx)
            all_subplan_root_dataset_ids.extend(subplan_root_dataset_ids)
            current_idx += 1

        # if current_idx is None, this is the first call, so we initialize it to 0
        if current_idx is None:
            current_idx = 0

        # compute this operator's unique logical_op_id
        this_unique_logical_op_id = f"{current_idx}-{self.logical_op_id}"

        # if this operator is a root dataset scan, update root_dataset_ids
        root_dataset_ids = []
        if isinstance(self.operator_set[0], MarshalAndScanDataOp):
            root_dataset_ids.append(self.operator_set[0].datasource.id)
        elif isinstance(self.operator_set[0], ContextScanOp):
            root_dataset_ids.append(self.operator_set[0].context.id)

        # update the root_dataset_ids_map for this operator
        root_dataset_ids_map[this_unique_logical_op_id] = root_dataset_ids + all_subplan_root_dataset_ids

        # return the current index and the upstream unique logical_op_ids for this operator
        return current_idx, root_dataset_ids_map[this_unique_logical_op_id]

    def get_root_dataset_ids(self, unique_logical_op_id: str) -> list[str]:
        """Return the list of root dataset ids which are upstream of this operator."""
        return self.unique_logical_op_id_to_root_dataset_ids[unique_logical_op_id]

    def _compute_source_unique_logical_op_ids_map(self, source_map: dict[str, list[str]], current_idx: int | None = None) -> tuple[int, str]:
        # get the topological index and logical_op_id pairs for all subplans' root operators
        subplan_topo_idx_op_id_pairs = []
        for subplan in self.subplans:
            current_idx, current_logical_op_id = subplan._compute_source_unique_logical_op_ids_map(source_map, current_idx)
            subplan_topo_idx_op_id_pairs.append((current_idx, current_logical_op_id))
            current_idx += 1

        # if current_idx is None, this is the first call, so we initialize it to 0
        if current_idx is None:
            current_idx = 0

        # compute this operator's unique logical_op_id
        this_unique_logical_op_id = f"{current_idx}-{self.logical_op_id}"

        # update the source_map for this operator
        source_map[this_unique_logical_op_id] = []
        for topo_idx, logical_op_id in subplan_topo_idx_op_id_pairs:
            unique_logical_op_id = f"{topo_idx}-{logical_op_id}"
            source_map[this_unique_logical_op_id].append(unique_logical_op_id)

        # return the current unique logical_op_id for this operator
        return current_idx, self.logical_op_id

    def get_source_unique_logical_op_ids(self, unique_logical_op_id: str) -> list[str]:
        """Return the list of unique logical_op_ids for the input(s) to this operator."""
        return self.unique_logical_op_id_to_source_logical_op_ids[unique_logical_op_id]



"""This file contains prompts for join operations."""

### BASE PROMPTS ###
JOIN_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with two data records and a join condition. Output TRUE if the two data records satisfy the join condition, and FALSE otherwise.

Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---

An example is shown below:
---
LEFT INPUT FIELDS:
{example_input_fields}

LEFT CONTEXT:
{{{example_context}}}{image_disclaimer}{audio_disclaimer}

RIGHT INPUT FIELDS:
{right_example_input_fields}

RIGHT CONTEXT:
{{{right_example_context}}}{right_image_disclaimer}{right_audio_disclaimer}

JOIN CONDITION: {example_join_condition}

Let's think step-by-step in order to evaluate the join condition.

REASONING: {example_reasoning}

ANSWER: TRUE
---
"""

JOIN_NO_REASONING_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with two data records and a join condition. Output TRUE if the two data records satisfy the join condition, and FALSE otherwise.

Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---

An example is shown below:
---
LEFT INPUT FIELDS:
{example_input_fields}

LEFT CONTEXT:
{{{example_context}}}{image_disclaimer}{audio_disclaimer}

RIGHT INPUT FIELDS:
{right_example_input_fields}

RIGHT CONTEXT:
{{{right_example_context}}}{right_image_disclaimer}{right_audio_disclaimer}

JOIN CONDITION: {example_join_condition}

ANSWER: TRUE
---
"""

JOIN_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with two data records and a join condition. Output TRUE if the two data records satisfy the join condition, and FALSE otherwise.
{desc_section}
Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---
---
LEFT INPUT FIELDS:
{input_fields_desc}

LEFT CONTEXT:
{context}<<image-audio-placeholder>>

RIGHT INPUT FIELDS:
{right_input_fields_desc}

RIGHT CONTEXT:
{right_context}<<right-image-audio-placeholder>>

JOIN CONDITION: {join_condition}

Let's think step-by-step in order to evaluate the join condition.

REASONING: """

JOIN_NO_REASONING_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with two data records and a join condition. Output TRUE if the two data records satisfy the join condition, and FALSE otherwise.
{desc_section}
Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---
---
LEFT INPUT FIELDS:
{input_fields_desc}

LEFT CONTEXT:
{context}<<image-audio-placeholder>>

RIGHT INPUT FIELDS:
{right_input_fields_desc}

RIGHT CONTEXT:
{right_context}<<right-image-audio-placeholder>>

JOIN CONDITION: {join_condition}

ANSWER: """
